"""Contains sklearn baseline models: ElasticNet, RandomForest, SVM."""

import numpy as np
from sklearn.ensemble import HistGradientBoostingRegressor, RandomForestRegressor
from sklearn.linear_model import ElasticNet, Lasso, Ridge
from sklearn.svm import SVR

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset
from drevalpy.models.drp_model import DRPModel

from ..utils import load_and_select_gene_features, load_drug_fingerprint_features


class SklearnModel(DRPModel):
    """Parent class that contains the common methods for the sklearn models."""

    cell_line_views = ["gene_expression"]
    drug_views = ["fingerprints"]

    def __init__(self):
        """
        Initializes the model.

        Sets the model to None, which is initialized in the build_model method to the respective sklearn model.
        """
        super().__init__()
        self.model = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError("get_model_name method has to be implemented in the child class.")

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Custom hyperparameters for the model, have to be defined in the child class.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError("build_model method has to be implemented in the child class.")

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the model.

        The number of features is the number of genes + the number of fingerprints.
        :param output: training dataset containing the response output
        :param cell_line_input: training dataset containing gene expression data
        :param drug_input: training dataset containing fingerprints data
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        :raises ValueError: If drug_input is None.
        """
        if drug_input is None:
            raise ValueError("drug_input (fingerprints) is required for the sklearn models.")

        x = self.get_concatenated_features(
            cell_line_view=self.cell_line_views[0],
            drug_view=self.drug_views[0],
            cell_line_ids_output=output.cell_line_ids,
            drug_ids_output=output.drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        self.model.fit(x, output.response)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the response for the given input.

        :param drug_ids: drug ids
        :param cell_line_ids: cell line ids
        :param drug_input: drug input
        :param cell_line_input: cell line input
        :returns: predicted drug response
        :raises ValueError: If drug_input is not None.
        """
        if drug_input is None:
            raise ValueError("drug_input (fingerprints) is required.")

        x = self.get_concatenated_features(
            cell_line_view=self.cell_line_views[0],
            drug_view=self.drug_views[0],
            cell_line_ids_output=cell_line_ids,
            drug_ids_output=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        return self.model.predict(x)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: Path to the gene expression and landmark genes
        :param dataset_name: Name of the dataset
        :returns: FeatureDataset containing the cell line gene expression features, filtered through the landmark genes
        """
        return load_and_select_gene_features(
            feature_type="gene_expression",
            gene_list="landmark_genes",
            data_path=data_path,
            dataset_name=dataset_name,
        )

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset | None:
        """
        Load the drug features, in this case the fingerprints.

        :param data_path: Path to the data
        :param dataset_name: Name of the dataset
        :returns: FeatureDataset containing the drug fingerprints
        """
        return load_drug_fingerprint_features(data_path, dataset_name, fill_na=True)


class ElasticNetModel(SklearnModel):
    """ElasticNet model for drug response prediction."""

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: ElasticNet
        """
        return "ElasticNet"

    def build_model(self, hyperparameters: dict):
        """
        Builds the ElasticNet model from hyperparameters.

        :param hyperparameters: Contains L1 ratio and alpha.
        """
        if hyperparameters["l1_ratio"] == 0.0:
            self.model = Ridge(alpha=hyperparameters["alpha"])
        elif hyperparameters["l1_ratio"] == 1.0:
            self.model = Lasso(alpha=hyperparameters["alpha"])
        else:
            self.model = ElasticNet(
                alpha=hyperparameters["alpha"],
                l1_ratio=hyperparameters["l1_ratio"],
            )


class RandomForest(SklearnModel):
    """RandomForest model for drug response prediction."""

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: RandomForest
        """
        return "RandomForest"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Hyperparameters for the model. Contains n_estimators, criterion, max_samples,
            and n_jobs.
        """
        if hyperparameters["max_depth"] == "None":
            hyperparameters["max_depth"] = None
        self.model = RandomForestRegressor(
            n_estimators=hyperparameters["n_estimators"],
            criterion=hyperparameters["criterion"],
            max_samples=hyperparameters["max_samples"],
            n_jobs=hyperparameters["n_jobs"],
        )


class SVMRegressor(SklearnModel):
    """SVM model for drug response prediction."""

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: SVR (Support Vector Regressor)
        """
        return "SVR"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Hyperparameters for the model. Contains kernel, C, epsilon, and max_iter.
        """
        self.model = SVR(
            kernel=hyperparameters["kernel"],
            C=hyperparameters["C"],
            epsilon=hyperparameters["epsilon"],
            max_iter=hyperparameters["max_iter"],
        )


class GradientBoosting(SklearnModel):
    """Gradient Boosting model for drug response prediction."""

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: GradientBoosting
        """
        return "GradientBoosting"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Hyperparameters for the model. Contains n_estimators, learning_rate, max_depth,
            and subsample
        """
        if hyperparameters["max_depth"] == "None":
            hyperparameters["max_depth"] = None
        self.model = HistGradientBoostingRegressor(
            max_iter=hyperparameters.get("max_iter", 100),
            learning_rate=hyperparameters.get("learning_rate", 0.1),
            max_depth=hyperparameters.get("max_depth", 3),
        )


class ProteomicsRandomForest(RandomForest):
    """RandomForest model for drug response prediction using proteomics data."""

    cell_line_views = ["proteomics"]

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: ProteomicsRandomForest
        """
        return "ProteomicsRandomForest"

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: Path to the gene expression and landmark genes
        :param dataset_name: Name of the dataset
        :returns: FeatureDataset containing the cell line proteomics features, filtered through the landmark genes
        """
        return load_and_select_gene_features(
            feature_type="proteomics",
            gene_list=None,
            data_path=data_path,
            dataset_name=dataset_name,
        )



"""Module for generating evaluation tables for cross-study drug response prediction."""

import os
import pathlib
from io import TextIOWrapper

import pandas as pd
import plotly.graph_objects as go

from ..pipeline_function import pipeline_function


class CrossStudyTables:
    """Generate evaluation tables for cross-study drug response prediction."""

    def __init__(self, evaluation_metrics: pd.DataFrame, path_data: pathlib.Path):
        """
        Initialize the CrossStudyTables object.

        :param evaluation_metrics: eval metrics dataframe.
        :param path_data: Path to data directory (used for context or extensions).
        """
        self.evaluation_metrics = evaluation_metrics
        self.path_data = path_data

        self.figures: dict[str, go.Figure] = {}
        cross_study_settings = evaluation_metrics[
            evaluation_metrics.rand_setting.str.contains("cross-study-")
        ].rand_setting.unique()
        self.cross_study_datasets = [setting.split("cross-study-")[1] for setting in cross_study_settings]

        evaluation_metrics = evaluation_metrics[evaluation_metrics.rand_setting.isin(cross_study_settings)]

        self.mean_metrics = []
        self.std_metrics = []
        for dataset in self.cross_study_datasets:
            evaluation_metrics_dataset = evaluation_metrics[
                evaluation_metrics.rand_setting.str.contains(f"cross-study-{dataset}")
            ]
            evaluation_metrics_group = [s.split("_split_")[0] for s in evaluation_metrics_dataset.index]
            metrics = [
                "MSE",
                "RMSE",
                "MAE",
                "R^2",
                "Pearson",
                "Spearman",
                "Kendall",
                "Pearson: normalized",
                "Spearman: normalized",
                "Kendall: normalized",
                "R^2: normalized",
            ]
            grouped = evaluation_metrics_dataset[metrics].groupby(evaluation_metrics_group)
            mean = grouped.mean()
            std = grouped.std()
            # sort by lowest MSE
            mean = mean.sort_values(by="MSE")
            std = std.loc[mean.index]

            mean.index = [s.split("_cross-study")[0] for s in mean.index]
            std.index = mean.index
            self.mean_metrics.append(mean)
            self.std_metrics.append(std)

    def draw(self):
        """Create and store Plotly table figures sorted by MSE."""
        for dataset_name, mean_df, std_df in zip(self.cross_study_datasets, self.mean_metrics, self.std_metrics):

            formatted_data = mean_df.map(lambda x: f"{x:.3f}") + " ± " + std_df.map(lambda x: f"{x:.3f}")

            fig = go.Figure(
                data=[
                    go.Table(
                        header=dict(
                            values=["Model"] + list(formatted_data.columns), fill_color="lightgrey", align="left"
                        ),
                        cells=dict(
                            values=[formatted_data.index]
                            + [formatted_data[col].values for col in formatted_data.columns],
                            fill_color="white",
                            align="left",
                        ),
                    )
                ]
            )
            fig.update_layout(title_text=f"Evaluation Metrics for Cross-Study Predictions to {dataset_name}")
            self.figures[dataset_name] = fig

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str):
        """
        Generate and save HTML tables for each cross-study dataset.

        :param out_prefix: Directory to save output files.
        :param out_suffix: Suffix to append to each output filename.
        """
        os.makedirs(out_prefix, exist_ok=True)
        self.draw()
        for dataset_name, fig in self.figures.items():
            filename = f"{out_prefix}/table_cross_study_{dataset_name}_{out_suffix}.html"
            fig.write_html(filename, include_plotlyjs="embed", full_html=True)

    @staticmethod
    def write_to_html(test_mode: str, f: TextIOWrapper, files: list[str], prefix: str) -> TextIOWrapper:
        """
        Embed HTML table files into an open HTML file handle.

        :param test_mode: Substring to match filenames (e.g., 'lpo', 'lco').
        :param f: Open writable file handle to insert HTML blocks.
        :param files: List of filenames in the target directory.
        :param prefix: Path prefix to locate HTML table files.

        :return: Updated file handle with HTML blocks written in.
        """
        if prefix:
            prefix = os.path.join(prefix, "html_tables")
        os.makedirs(prefix, exist_ok=True)

        for file in files:
            if file.startswith("table_cross_study_") and file.endswith(".html") and test_mode in file:
                f.write(f'<iframe src="html_tables/{file}" width="100%" height="600" frameborder="0"></iframe>\n')
        return f



"""
Implements the naive predictor models.

The naive predictor models are simple models that predict the mean of the response values. The NaivePredictor
predicts the overall mean of the response, the NaiveCellLineMeanPredictor predicts the mean of the response per cell
line, and the NaiveDrugMeanPredictor predicts the mean of the response per drug.
The NaiveMeanEffectsPredictor predicts the response as the overall mean plus the cell line effect
plus the drug effect and should be the strongest naive baseline.

"""

import numpy as np

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset
from drevalpy.datasets.utils import CELL_LINE_IDENTIFIER, DRUG_IDENTIFIER, TISSUE_IDENTIFIER
from drevalpy.models.drp_model import DRPModel
from drevalpy.models.utils import load_cl_ids_from_csv, load_drug_ids_from_csv, load_tissues_from_csv, unique


class NaivePredictor(DRPModel):
    """Naive predictor model that predicts the overall mean of the response."""

    cell_line_views = [CELL_LINE_IDENTIFIER]
    drug_views = [DRUG_IDENTIFIER]

    def __init__(self):
        """
        Initializes the model.

        Sets the dataset mean to None, which is initialized in the train method.
        """
        super().__init__()
        self.dataset_mean = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: NaivePredictor
        """
        return "NaivePredictor"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters. Not needed for the NaivePredictor.

        :param hyperparameters: Hyperparameters for the model, not needed
        """
        pass

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Computes the overall mean of the output response values and saves them.

        :param output: training dataset containing the response output
        :param cell_line_input: not needed
        :param drug_input: not needed
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        """
        self.dataset_mean = np.mean(output.response)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the dataset mean for each drug-cell line combination.

        :param cell_line_ids: cell line ids
        :param drug_ids: not needed
        :param cell_line_input: not needed
        :param drug_input: not needed
        :return: array of the same length as the input cell line id containing the dataset mean
        """
        return np.full(cell_line_ids.shape[0], self.dataset_mean)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features, in this case the cell line ids.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset containing the cell line ids
        """
        return load_cl_ids_from_csv(data_path, dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features, in this case the drug ids.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset containing the drug ids
        """
        return load_drug_ids_from_csv(data_path, dataset_name)


class NaiveDrugMeanPredictor(DRPModel):
    """Naive predictor model that predicts the mean of the response per drug."""

    cell_line_views = [CELL_LINE_IDENTIFIER]
    drug_views = [DRUG_IDENTIFIER]

    def __init__(self):
        """
        Initializes the model.

        Drug means and dataset mean are set to None, which are initialized in the train method.
        """
        super().__init__()
        self.drug_means = None
        self.dataset_mean = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: NaiveDrugMeanPredictor
        """
        return "NaiveDrugMeanPredictor"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters. Not needed for the NaiveDrugMeanPredictor.

        :param hyperparameters: Hyperparameters for the model, not needed
        """
        pass

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "None",
    ) -> None:
        """
        Computes the mean per drug. If - later on - the drug is not in the training set, the overall mean is used.

        :param output: training dataset containing the response output
        :param cell_line_input: not needed
        :param drug_input: drug id
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        :raises ValueError: If drug_input is None
        """
        if drug_input is None:
            raise ValueError("drug_input (drug_id) is required for the NaiveDrugMeanPredictor.")
        drug_ids = drug_input.get_feature_matrix(view=DRUG_IDENTIFIER, identifiers=output.drug_ids)
        self.dataset_mean = np.mean(output.response)
        self.drug_means = {}

        for drug_response, drug_feature in zip(unique(output.drug_ids), unique(drug_ids), strict=True):
            responses_drug = output.response[drug_feature == output.drug_ids]
            if len(responses_drug) > 0:
                # prevent nan response
                self.drug_means[drug_response] = np.mean(responses_drug)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the drug mean for each drug-cell line combination.

        If the drug is not in the training set, the dataset mean is used.

        :param cell_line_ids: not needed
        :param drug_ids: drug ids
        :param cell_line_input: not needed
        :param drug_input: not needed
        :return: array of the same length as the input drug_id containing the drug mean
        """
        return np.array([self.predict_drug(drug) for drug in drug_ids])

    def predict_drug(self, drug_id: str):
        """
        Predicts the mean of the response for a given drug.

        If the drug is not in the training set, the dataset mean is used.

        :param drug_id: ID of the drug
        :return: predicted response
        """
        if drug_id in self.drug_means:
            return self.drug_means[drug_id]
        return self.dataset_mean

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: Path to the data.
        :param dataset_name: Name of the dataset.
        :return: FeatureDataset containing the cell line IDs.
        """
        return load_cl_ids_from_csv(data_path, dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features, in this case the drug ids.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset containing the drug ids
        """
        return load_drug_ids_from_csv(data_path, dataset_name)


class NaiveCellLineMeanPredictor(DRPModel):
    """Naive predictor model that predicts the mean of the response per cell line."""

    cell_line_views = [CELL_LINE_IDENTIFIER]
    drug_views = [DRUG_IDENTIFIER]

    def __init__(self):
        """
        Initializes the model.

        Cell line means and dataset mean are set to None, which are initialized in the train method.
        """
        super().__init__()
        self.cell_line_means = None
        self.dataset_mean = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: NaiveCellLineMeanPredictor
        """
        return "NaiveCellLineMeanPredictor"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters. Not needed for the NaiveCellLineMeanPredictor.

        :param hyperparameters: not needed
        """
        pass

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "None",
    ) -> None:
        """
        Computes the mean per cell line.

        If - later on - the cell line is not in the training set, the overall mean is used.
        :param output: training dataset containing the response output
        :param cell_line_input: cell line inputs
        :param drug_input: not needed
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        """
        cell_line_ids = cell_line_input.get_feature_matrix(view=CELL_LINE_IDENTIFIER, identifiers=output.cell_line_ids)
        self.dataset_mean = np.mean(output.response)
        self.cell_line_means = {}

        for cell_line_response, cell_line_feature in zip(
            unique(output.cell_line_ids), unique(cell_line_ids), strict=True
        ):
            responses_cl = output.response[cell_line_feature == output.cell_line_ids]
            if len(responses_cl) > 0:
                # prevent nan response
                self.cell_line_means[cell_line_response] = np.mean(responses_cl)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the cell line mean for each drug-cell line combination.

        If the cell line is not in the training set, the dataset mean is used.

        :param cell_line_ids: cell line ids
        :param drug_ids: not needed
        :param cell_line_input: not needed
        :param drug_input: not needed
        :return: array of the same length as the input cell_line_id containing the cell line mean
        """
        return np.array([self.predict_cl(cl) for cl in cell_line_ids])

    def predict_cl(self, cl_id: str) -> float:
        """
        Predicts the mean of the response for a given cell line.

        If the cell line is not in the training set, the dataset mean is used.
        :param cl_id: Cell line ID
        :return: predicted response
        """
        if cl_id in self.cell_line_means:
            return self.cell_line_means[cl_id]
        return self.dataset_mean

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features, in this case the cell line ids.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset containing the cell line ids
        """
        return load_cl_ids_from_csv(data_path, dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features.

        :param data_path: Path to the data.
        :param dataset_name: Name of the dataset.
        :return: FeatureDataset containing the drug IDs.
        """
        return load_drug_ids_from_csv(data_path, dataset_name)


class NaiveTissueMeanPredictor(DRPModel):
    """Naive predictor model that predicts the mean of the response per tissue."""

    cell_line_views = [TISSUE_IDENTIFIER]
    drug_views = []

    def __init__(self):
        """
        Initializes the model.

        Tissue means and dataset mean are set to None, which are initialized in the train method.
        """
        super().__init__()
        self.tissue_means = None
        self.dataset_mean = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: NaiveTissueMeanPredictor
        """
        return "NaiveTissueMeanPredictor"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters. Not needed for the NaiveTissueMeanPredictor.

        :param hyperparameters: Hyperparameters for the model, not needed
        """
        pass

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "None",
    ) -> None:
        """
        Computes the mean per tissue. Falls back to the overall mean for unknown tissues.

        :param output: training dataset with `.response` and `.tissue`
        :param cell_line_input: not needed
        :param drug_input: not needed
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        :raises ValueError: If tissue information is missing in the output dataset.
        """
        self.dataset_mean = np.mean(output.response)
        self.tissue_means = {}

        if output.tissue is None:
            raise ValueError("Tissue information is missing in the output dataset.")
        for tissue in np.unique(output.tissue):
            mask = output.tissue == tissue
            responses = output.response[mask]
            if len(responses) > 0:
                self.tissue_means[tissue] = np.mean(responses)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the tissue mean for each drug-cell line combination.

        If the tissue is not in the training set, the dataset mean is used.

        :param cell_line_ids: cell line ids
        :param drug_ids: not needed
        :param cell_line_input: tissue features
        :param drug_input: not needed
        :return: array of the same length as the input cell_line_id containing the tissue mean
        """
        tissues = cell_line_input.get_feature_matrix(view=TISSUE_IDENTIFIER, identifiers=cell_line_ids)
        preds = []
        for tissue in tissues:
            key = tissue.item() if isinstance(tissue, np.ndarray) else tissue
            preds.append(self.tissue_means.get(key, self.dataset_mean))
        return np.array(preds)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features, in this case the tissue annotations.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset containing the tissue ids
        """
        return load_tissues_from_csv(data_path, dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features.

        :param data_path: Path to the data.
        :param dataset_name: Name of the dataset.
        :return: FeatureDataset containing the drug IDs.
        """
        return load_drug_ids_from_csv(data_path, dataset_name)


class NaiveMeanEffectsPredictor(DRPModel):
    """
    ANOVA-like predictor model.

    Predicts the response as:
    response = overall_mean + cell_line_effect + drug_effect.

    Here:
        - cell_line_effect = (cell line mean - overall_mean)
        - drug_effect = (drug mean - overall_mean)

    This formulation ensures that the overall mean is not counted twice.
    """

    cell_line_views = [CELL_LINE_IDENTIFIER]
    drug_views = [DRUG_IDENTIFIER]

    def __init__(self):
        """
        Initializes the NaiveMeanEffectsPredictor model.

        The overall dataset mean, cell line effects, and drug effects are initialized to None
        and empty dictionaries, respectively.
        """
        super().__init__()
        self.dataset_mean = None
        self.cell_line_effects = {}
        self.drug_effects = {}

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the name of the model.

        :return: The name of the model as a string.
        """
        return "NaiveMeanEffectsPredictor"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model.

        This model does not require any hyperparameter tuning.

        :param hyperparameters: Dictionary of hyperparameters (not used).
        """
        pass

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains with overall mean, cell line effects, and drug effects.

        :param output: Training dataset containing the response output.
        :param cell_line_input: Feature dataset containing cell line IDs.
        :param drug_input: Feature dataset containing drug IDs. Must not be None.
        :param output_earlystopping: Not used.
        :param model_checkpoint_dir: Not used.
        :raises ValueError: If drug_input is None.
        """
        if drug_input is None:
            raise ValueError("drug_input (drug_id) is required for ANOVAPredictor.")

        # Compute the overall mean response.
        self.dataset_mean = np.mean(output.response)

        # Obtain cell line features.
        cell_line_ids = cell_line_input.get_feature_matrix(view=CELL_LINE_IDENTIFIER, identifiers=output.cell_line_ids)
        cell_line_means = {}
        for cl_output, cl_feature in zip(unique(output.cell_line_ids), unique(cell_line_ids), strict=True):
            responses_cl = output.response[cl_feature == output.cell_line_ids]
            if len(responses_cl) > 0:
                cell_line_means[cl_output] = np.mean(responses_cl)

        # Obtain drug features.
        drug_ids = drug_input.get_feature_matrix(view=DRUG_IDENTIFIER, identifiers=output.drug_ids)
        drug_means = {}
        for drug_output, drug_feature in zip(unique(output.drug_ids), unique(drug_ids), strict=True):
            responses_drug = output.response[drug_feature == output.drug_ids]
            if len(responses_drug) > 0:
                drug_means[drug_output] = np.mean(responses_drug)

        # Compute the effects as deviations from the overall mean.
        self.cell_line_effects = {cl: (mean - self.dataset_mean) for cl, mean in cell_line_means.items()}
        self.drug_effects = {drug: (mean - self.dataset_mean) for drug, mean in drug_means.items()}

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts responses for given cell line and drug pairs.

        The prediction is computed as:
            prediction = overall_mean + cell_line_effect + drug_effect

        If a cell line or drug has not been seen during training, their effect is set to zero.

        :param cell_line_ids: Array of cell line IDs.
        :param drug_ids: Array of drug IDs.
        :param cell_line_input: Not used.
        :param drug_input: Not used.
        :return: NumPy array of predicted responses.
        """
        predictions = []
        for cl, drug in zip(cell_line_ids, drug_ids):
            effect_cl = self.cell_line_effects.get(cl, 0)
            effect_drug = self.drug_effects.get(drug, 0)
            # ANOVA-based prediction: overall mean + cell line effect + drug effect.
            predictions.append(self.dataset_mean + effect_cl + effect_drug)
        return np.array(predictions)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: Path to the data.
        :param dataset_name: Name of the dataset.
        :return: FeatureDataset containing the cell line IDs.
        """
        return load_cl_ids_from_csv(data_path, dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features.

        :param data_path: Path to the data.
        :param dataset_name: Name of the dataset.
        :return: FeatureDataset containing the drug IDs.
        """
        return load_drug_ids_from_csv(data_path, dataset_name)



"""
Includes functions to load and process the DIPK dataset.

get_data: Creates a list of dictionaries with drug and cell line features.
CollateFn: Class to collate the DataLoader batches.
DIPKDataset: Dataset class for the DIPK model.

"""

import os
from abc import ABC

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset

from drevalpy.datasets.dataset import FeatureDataset


def load_bionic_features(data_path: str, dataset_name: str, gene_add_num: int = 512) -> FeatureDataset:
    """
    Load biological network (BIONIC) features for DIPK.

    :param data_path: Path to the data, e.g., "data/"
    :param dataset_name: Name of the dataset, e.g., GDSC2
    :param gene_add_num: Number of genes to add to the feature set
    :returns: FeatureDataset with gene expression and biological network features
    """
    # Load gene expression dataset
    gene_expression_path = os.path.join(data_path, dataset_name, "gene_expression.csv")
    gene_expression = pd.read_csv(gene_expression_path)
    expression_dict = gene_expression.set_index("cell_line_name").drop("cellosaurus_id", axis=1).T.to_dict()

    # Load gene list and PPI features
    gene_list_path = os.path.join(data_path, dataset_name, "DIPK_features", "gene_list_sel.txt")
    with open(gene_list_path, encoding="gbk") as f:
        gene_list = {line.strip() for line in f}

    ppi_path = os.path.join(data_path, dataset_name, "DIPK_features", "human_ppi_features.tsv")
    dataset = pd.read_csv(ppi_path, index_col=0, sep="\t")

    # Ensure BIONIC dictionary uses gene names directly
    bionic_gene_dict = {gene: dataset.loc[gene].values for gene in gene_list if gene in dataset.index}

    # Compute BIONIC features
    bionic_feature_dict = {}
    for cell_line, expressions in expression_dict.items():
        # Sort genes based on descending expression values
        sorted_genes = sorted(expressions.items(), key=lambda x: -x[1])
        top_genes = [gene for gene, _ in sorted_genes[:gene_add_num]]

        # Aggregate BIONIC features for selected genes
        selected_features = [bionic_gene_dict[gene] for gene in top_genes if gene in bionic_gene_dict]
        if selected_features:
            aggregated_feature = np.mean(selected_features, axis=0)
        else:
            # Handle case where no features are found (padding with zeros)
            aggregated_feature = np.zeros(next(iter(bionic_gene_dict.values())).shape)

        bionic_feature_dict[cell_line] = aggregated_feature

    feature_data = {cell_line: {"bionic_features": features} for cell_line, features in bionic_feature_dict.items()}
    return FeatureDataset(features=feature_data)


def get_data(
    cell_ids: np.ndarray,
    drug_ids: np.ndarray,
    cell_line_features: FeatureDataset,
    drug_features: FeatureDataset,
    ic50: np.ndarray | None = None,
) -> list:
    """
    Prepare data samples for training or prediction.

    Each sample includes:
    - Drug features (e.g., molecular embeddings).
    - Cell line features (gene expression and bionic_features).
    - Optional IC50 response values for supervised tasks.

    :param cell_ids: IDs of the cell lines from the dataset.
    :param drug_ids: IDs of the drugs from the dataset.
    :param cell_line_features: Input features associated with the cell lines.
    :param drug_features: Input features associated with the drugs.
    :param ic50: (Optional) Response values (e.g., IC50) to associate with samples.
    :return: List of dictionaries, each containing drug and cell line features, with optional IC50.
    """
    data_list = []
    for i in range(len(cell_ids)):
        drug_id = str(drug_ids[i])
        cell_id = str(cell_ids[i])
        drug_tensor = torch.tensor(drug_features.features[drug_id]["molgnet_features"], dtype=torch.float32)
        gene_expression = torch.tensor(cell_line_features.features[cell_id]["gene_expression"], dtype=torch.float32)
        bionic_features = torch.tensor(cell_line_features.features[cell_id]["bionic_features"], dtype=torch.float32)

        sample = {
            "molgnet_features": drug_tensor,
            "gene_expression": gene_expression,
            "bionic_features": bionic_features,
        }
        if ic50 is not None:
            sample["ic50"] = torch.tensor([ic50[i]], dtype=torch.float32)

        data_list.append(sample)

    return data_list


class CollateFn:
    """Collate function for the DataLoader, either for training or testing."""

    def __init__(self, train=True):
        """
        Initialize the CollateFn.

        :param train: indicates whether the DataLoader is used for training
        """
        self.train = train

    def __call__(self, batch):
        """
        Collate the batch.

        :param batch: batch of feature dictionaries
        :returns: collated node features, gene features, bionic features, and (optional) IC50 values
        """
        # Find the max number of atoms (nodes) in the batch for molgnet_features padding
        max_atoms_molgnet = max([sample["molgnet_features"].size(0) for sample in batch])

        # Pad molgnet_features to match the maximum number of atoms
        padded_molgnet_features = []
        molgnet_mask = []

        for sample in batch:
            num_atoms = sample["molgnet_features"].size(0)
            padding_size = max_atoms_molgnet - num_atoms

            # Pad molgnet_features
            padded_features = torch.cat(
                [sample["molgnet_features"], torch.zeros(padding_size, sample["molgnet_features"].size(1))], dim=0
            )
            padded_molgnet_features.append(padded_features)

            # Create a mask where valid atom features are True and padded ones are False
            mask = torch.cat(
                [torch.ones(num_atoms, dtype=torch.bool), torch.zeros(padding_size, dtype=torch.bool)], dim=0
            )
            molgnet_mask.append(mask)

        # Stack the padded molgnet features into a single tensor
        molgnet_features = torch.stack(padded_molgnet_features)
        molgnet_mask = torch.stack(molgnet_mask)

        # Collate other features
        gene_features = torch.stack([sample["gene_expression"] for sample in batch])
        bionic_features = torch.stack([sample["bionic_features"] for sample in batch])

        if self.train:
            ic50_values = torch.stack([sample["ic50"] for sample in batch])
            # Return a dictionary with all features
            return {
                "molgnet_features": molgnet_features,
                "gene_features": gene_features,
                "bionic_features": bionic_features,
                "ic50_values": ic50_values,
                "molgnet_mask": molgnet_mask,
            }
        else:
            # Return a dictionary without ic50_values for inference
            return {
                "molgnet_features": molgnet_features,
                "gene_features": gene_features,
                "bionic_features": bionic_features,
                "molgnet_mask": molgnet_mask,
            }


class DIPKDataset(Dataset, ABC):
    """Dataset of graphs from get_data."""

    def __init__(self, samples):
        """
        Initialize the GraphDataset.

        :param samples: list
        """
        super().__init__()
        self._samples = samples

    def __getitem__(self, idx):
        """
        Get the sample at index idx.

        :param idx: index
        :returns: sample
        """
        sample = self._samples[idx]
        return sample

    def __len__(self) -> int:
        """
        Get the number of graphs in the dataset.

        :return: number of samples
        """
        return len(self._samples)



"""
Utility functions for the MOLIR model.

Original authors of MOLI: Sharifi-Noghabi et al. (2019, 10.1093/bioinformatics/btz318)
Code adapted from: Hauptmann et al. (2023, 10.1186/s12859-023-05166-7),
https://github.com/kramerlab/Multi-Omics_analysis
"""

import os
import random
import secrets

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import EarlyStopping, TQDMProgressBar
from torch import nn
from torch.utils.data import DataLoader, Dataset

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset
from drevalpy.models.drp_model import DRPModel


class RegressionDataset(Dataset):
    """Dataset for regression tasks for the data loader."""

    def __init__(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
    ) -> None:
        """
        Initializes the dataset by setting the output and the cell line input.

        :param output: drug response dataset
        :param cell_line_input: omics features of the cell lines
        """
        self.output = output
        self.cell_line_input = cell_line_input

    def __getitem__(self, idx: int) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.float32]:
        """
        Overwrites the getitem method.

        :param idx: index of the sample
        :returns: gene expression, mutations, copy number variation, and response of the sample as numpy arrays
        """
        response: np.float32 = np.float32(self.output.response[idx])

        cell_line_id = str(self.output.cell_line_ids[idx])
        gene_expression: np.ndarray = self.cell_line_input.features[cell_line_id]["gene_expression"].astype(np.float32)
        mutations: np.ndarray = self.cell_line_input.features[cell_line_id]["mutations"].astype(np.float32)
        copy_number: np.ndarray = self.cell_line_input.features[cell_line_id]["copy_number_variation_gistic"].astype(
            np.float32
        )

        return gene_expression, mutations, copy_number, response

    def __len__(self) -> int:
        """
        Overwrites the len method.

        :returns: number of samples in the dataset
        """
        return len(self.output.response)


def generate_triplets_indices(
    y: np.ndarray,
    positive_range: float,
    negative_range: float,
    random_seed: int | None = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Generates triplets for the MOLIR model.

    The positive and negative range are determined by the standard deviation of the response values. A sample is
    considered positive if its response value is within the positive range of the label. The positive range is ±10%
    of the standard deviation of all response values. A sample is considered negative if its response value is at
    least one standard deviation away from the response value of the sample.
    :param y: response values
    :param positive_range: positive range for the triplet loss
    :param negative_range: negative range for the triplet loss
    :param random_seed: random seed for reproducibility
    :returns: positive and negative sample indices for each sample
    """
    if random_seed is not None:
        random.seed(random_seed)
        np.random.seed(random_seed)
    positive_sample_indices = []
    negative_sample_indices = []
    # Iterate over each label in the dataset
    for idx_current_label, current_label in enumerate(y):
        positive_class_indices = _get_positive_class_indices(current_label, idx_current_label, y, positive_range)
        positive_sample_idx = np.random.choice(positive_class_indices, 1)[0]
        negative_class_indices = _get_negative_class_indices(current_label, y, negative_range)
        negative_sample_idx = np.random.choice(negative_class_indices, 1)[0]
        positive_sample_indices.append(positive_sample_idx)
        negative_sample_indices.append(negative_sample_idx)
    return np.array(positive_sample_indices), np.array(negative_sample_indices)


def _get_positive_class_indices(label: np.float32, idx_label: int, y: np.ndarray, positive_range: float) -> np.ndarray:
    """
    Find the samples that are within the positive range of the label except the label itself.

    If there is no similar sample within the positive range, the method returns the closest sample to the label.
    :param label: response of interest
    :param idx_label: index of the response of interest
    :param y: all responses
    :param positive_range: 0.1 * the standard deviation of all training responses
    :returns: indices of the samples that can be considered positive examples (=similar to the response of interest)
    """
    indices_similar_samples = np.where(np.logical_and(label - positive_range <= y, y <= label + positive_range))[0]
    indices_similar_samples = np.delete(indices_similar_samples, np.where(indices_similar_samples == idx_label))
    if len(indices_similar_samples) == 0 and len(y) > 1:
        # return the closest samples to the label except the label itself
        indices_similar_samples = np.array([np.argsort(np.abs(y - label))[1]])
    elif len(indices_similar_samples) == 0:
        # this happens only when there is just one sample in the validation dataset
        indices_similar_samples = np.where(y == label)[0]
    return indices_similar_samples


def _get_negative_class_indices(label: np.float32, y: np.ndarray, negative_range: float) -> np.ndarray:
    """
    Finds dissimilar samples to the label.

    If there is no dissimilar sample within the negative range, the method returns the sample that is the furthest away.
    :param label: reponse of interest
    :param y: all responses
    :param negative_range: 1 * the standard deviation of all training responses
    :returns: indices of the samples that can be considered negative examples (=dissimilar to the response of interest)
    """
    dissimilar_samples = np.where(np.logical_or(label - negative_range >= y, y >= label + negative_range))[0]
    if len(dissimilar_samples) == 0:
        # return the sample that is the furthest away from the label
        dissimilar_samples = np.argsort(np.abs(y - label))[-1:]
    return dissimilar_samples


def make_ranges(output: DrugResponseDataset) -> tuple[float, float]:
    """
    Compute the positive and negative range for the triplet loss.

    :param output: drug response dataset
    :returns: positive and negative range for the triplet loss
    """
    positive_range = float(np.std(output.response) * 0.1)
    negative_range = float(np.std(output.response))
    return positive_range, negative_range


def create_dataset_and_loaders(
    batch_size: int,
    output_train: DrugResponseDataset,
    cell_line_input: FeatureDataset,
    output_earlystopping: DrugResponseDataset | None = None,
) -> tuple[DataLoader, DataLoader | None]:
    """
    Creates the RegressionDataset (torch Dataset) and the DataLoader for the training and validation data.

    :param batch_size: specified batch size
    :param output_train: response values for the training data
    :param cell_line_input: omic input features of the cell lines
    :param output_earlystopping: early stopping dataset
    :returns: training and validation data loaders
    """
    train_dataset = RegressionDataset(output_train, cell_line_input)
    train_loader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=1 if os.name == "nt" else 4,  # multiprocessing on Windows is not supported
        persistent_workers=True,
        drop_last=True,  # avoids batch norm errors if last batch < batch_size
    )

    val_loader = None
    if output_earlystopping is not None:
        val_dataset = RegressionDataset(
            output=output_earlystopping,
            cell_line_input=cell_line_input,
        )
        val_loader = DataLoader(
            val_dataset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=1,
            persistent_workers=True,
        )
    return train_loader, val_loader


def get_dimensions_of_omics_data(cell_line_input: FeatureDataset) -> tuple[int, int, int]:
    """
    Determines the dimensions of the omics data for the creation of the input layers.

    :param cell_line_input: omic input features of the cell lines
    :returns: dimensions of the gene expression, mutations, and copy number variation data
    """
    first_item = next(iter(cell_line_input.features.values()))
    dim_gex = first_item["gene_expression"].shape[0]
    dim_mut = first_item["mutations"].shape[0]
    dim_cnv = first_item["copy_number_variation_gistic"].shape[0]
    return dim_gex, dim_mut, dim_cnv


def filter_and_sort_omics(
    model: DRPModel,  # MOLIR or SuperFELTR
    gene_expression: np.ndarray,
    mutations: np.ndarray,
    cnvs: np.ndarray,
    cell_line_input: FeatureDataset,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Filters out features that were not present during training and imputes missing features with zeros.

    This is necessary because the feature order might have changed or more features are available (cross-study setting).

    :param model: either MOLIR or SuperFELTR self
    :param gene_expression: new gene expression data from which to predict
    :param mutations: new mutation data from which to predict
    :param cnvs: new copy number variation data from which to predict
    :param cell_line_input: needed for meta information (feature names)
    :return: filtered and sorted gene expression, mutations, and copy number variation data
    """
    for key, features in {
        "gene_expression": model.gene_expression_features,  # type: ignore
        "mutations": model.mutations_features,  # type: ignore
        "copy_number_variation_gistic": model.copy_number_variation_features,  # type: ignore
    }.items():
        if key == "gene_expression":
            values = gene_expression
        elif key == "mutations":
            values = mutations
        else:
            values = cnvs
        if values.shape[1] != len(features):
            new_value = np.zeros((values.shape[0], len(features)))
            lookup_table = {feature: i for i, feature in enumerate(cell_line_input.meta_info[key])}
            for i, feature in enumerate(features):
                if feature in lookup_table:
                    new_value[:, i] = values[:, lookup_table[feature]]
            if key == "gene_expression":
                gene_expression = new_value
            elif key == "mutations":
                mutations = new_value
            else:
                cnvs = new_value
    return gene_expression, mutations, cnvs


def select_features_for_view(
    view: str,  # "gene_expression", "mutations", or "copy_number_variation_gistic"
    cell_line_input: FeatureDataset,
    output: DrugResponseDataset,
):
    """
    Selects the top 1000 features with the highest variance for the omics data.

    :param view: either "gene_expression", "mutations", or "copy_number_variation_gistic"
    :param cell_line_input: the omics data of the cell lines
    :param output: the training dataset containing the response output
    :return: the modified cell line input with the top 1000 features with the highest variance
    """
    train_features = np.vstack(
        [cell_line_input.features[identifier][view] for identifier in np.unique(output.cell_line_ids)]
    )
    variances = np.var(train_features, axis=0)
    mask = np.zeros(len(variances), dtype=bool)
    mask[np.argsort(variances)[::-1][:1000]] = True
    cell_line_input.meta_info[view] = cell_line_input.meta_info[view][mask]
    for identifier in cell_line_input.features:
        cell_line_input.features[identifier][view] = cell_line_input.features[identifier][view][mask]
    return cell_line_input


class MOLIEncoder(nn.Module):
    """
    Encoders of the MOLIR model, which is identical to the encoders of the original MOLI model.

    The MOLIR model has three encoders for the gene expression, mutations, and copy number variation data which are
    trained together.
    """

    def __init__(self, input_size: int, output_size: int, dropout_rate: float) -> None:
        """
        Initializes the encoder for the MOLIR model.

        :param input_size: input size determined by feature selection.
        :param output_size: output size of the encoder, set as hyperparameter.
        :param dropout_rate: dropout rate for regularization, set as hyperparameter.
        """
        super().__init__()
        self.encode = nn.Sequential(
            nn.Linear(input_size, output_size),
            nn.ReLU(),
            nn.BatchNorm1d(output_size),
            nn.Dropout(dropout_rate),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the encoder.

        :param x: omic input features
        :returns: encoded omic features
        """
        return self.encode(x)


class MOLIRegressor(nn.Module):
    """
    Regressor of the MOLIR model.

    It is identical to the regressor of the original MOLI model, except for the omission of the final sigmoid
    activation function. After the three encoders, the encoded features are concatenated and fed into the regressor.
    """

    def __init__(self, input_size: int, dropout_rate: float) -> None:
        """
        Initializes the regressor for the MOLIR model.

        :param input_size: determined by the output sizes of the encoders.
        :param dropout_rate: set as hyperparameter.
        """
        super().__init__()
        self.regressor = nn.Sequential(nn.Linear(input_size, 1), nn.Dropout(dropout_rate))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the regressor.

        :param x: concatenated encoded features
        :returns: predicted drug response
        """
        return self.regressor(x)


class MOLIModel(pl.LightningModule):
    """
    PyTorch Lightning module for the MOLIR model.

    The architecture of the MOLIR model is identical to the MOLI model, except for the omission of the final sigmoid
    layer and the usage of a regression MSE loss instead of a binary cross-entropy loss. Additionally, early stopping is
    added instead of tuning the number of epochs as hyperparameter.
    """

    def __init__(
        self, hpams: dict[str, int | float], input_dim_expr: int, input_dim_mut: int, input_dim_cnv: int
    ) -> None:
        """
        Initializes the MOLIR model.

        The MOLIR model uses a combined loss function of a triplet margin loss for the concatenated representation
        and an MSE loss for the regression loss.

        :param hpams: includes mini_batch, layer dimensions (h_dim1, h_dim2, h_dim3), learning_rate, dropout_rate,
            weight decay, gamma, epochs, and margin.
        :param input_dim_expr: determined by the feature selection of the gene expression data.
        :param input_dim_mut: determined by dataset size
        :param input_dim_cnv: determined by dataset size
        """
        super().__init__()
        self.save_hyperparameters()

        self.mini_batch = int(hpams["mini_batch"])
        self.h_dim1 = int(hpams["h_dim1"])
        self.h_dim2 = int(hpams["h_dim2"])
        self.h_dim3 = int(hpams["h_dim3"])
        self.lr = hpams["learning_rate"]
        self.dropout_rate = hpams["dropout_rate"]
        self.weight_decay = hpams["weight_decay"]
        self.gamma = hpams["gamma"]
        self.epochs = int(hpams["epochs"])
        self.triplet_loss = nn.TripletMarginLoss(margin=hpams["margin"], p=2)
        self.regression_loss = nn.MSELoss()
        # Positive and Negative range for triplet loss, determined by the standard deviation of the training responses,
        # set in fit method
        self.positive_range = 1.0
        self.negative_range = 1.0
        # Checkpoint callback for early stopping, set in fit method
        self.checkpoint_callback: pl.callbacks.ModelCheckpoint | None = None

        self.expression_encoder = MOLIEncoder(input_dim_expr, self.h_dim1, self.dropout_rate)
        self.mutation_encoder = MOLIEncoder(input_dim_mut, self.h_dim2, self.dropout_rate)
        self.cna_encoder = MOLIEncoder(input_dim_cnv, self.h_dim3, self.dropout_rate)
        self.regressor = MOLIRegressor(self.h_dim1 + self.h_dim2 + self.h_dim3, self.dropout_rate)

    def fit(
        self,
        output_train: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        output_earlystopping: DrugResponseDataset | None = None,
        patience: int = 5,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the MOLIR model.

        First, the ranges for the triplet loss are determined using the standard deviation of the training responses.
        Then, the training and validation data loaders are created. The model is trained using the Lightning Trainer
        with an early stopping callback and patience of 5.
        :param output_train: training dataset containing the response output
        :param cell_line_input: feature dataset containing the omics data of the cell lines
        :param output_earlystopping: early stopping dataset
        :param patience: for early stopping
        :param model_checkpoint_dir: directory to save the model checkpoints
        """
        self.positive_range, self.negative_range = make_ranges(output_train)

        train_loader, val_loader = create_dataset_and_loaders(
            batch_size=self.mini_batch,
            output_train=output_train,
            cell_line_input=cell_line_input,
            output_earlystopping=output_earlystopping,
        )

        # Train the model
        monitor = "train_loss" if (val_loader is None) else "val_loss"

        early_stop_callback = EarlyStopping(monitor=monitor, mode="min", patience=patience)
        name = "version-" + "".join(
            [secrets.choice("0123456789abcdef") for _ in range(20)]
        )  # preventing conflicts of filenames
        self.checkpoint_callback = pl.callbacks.ModelCheckpoint(
            dirpath=os.path.join(model_checkpoint_dir, name),
            monitor=monitor,
            mode="min",
            save_top_k=1,
            save_weights_only=True,
        )

        # Initialize the Lightning trainer
        trainer = pl.Trainer(
            max_epochs=self.epochs,
            callbacks=[
                early_stop_callback,
                self.checkpoint_callback,
                TQDMProgressBar(refresh_rate=0),
            ],
            devices=1,
            enable_model_summary=False,
        )
        if val_loader is None:
            trainer.fit(self, train_loader)
        else:
            trainer.fit(self, train_loader, val_loader)
        # load best model
        if self.checkpoint_callback.best_model_path is not None:
            checkpoint = torch.load(self.checkpoint_callback.best_model_path, weights_only=True)  # noqa: S614
            self.load_state_dict(checkpoint["state_dict"])

    def predict(
        self,
        gene_expression: np.ndarray,
        mutations: np.ndarray,
        copy_number: np.ndarray,
    ) -> np.ndarray:
        """
        Perform prediction on given input data.

        If there was enough training data to train the model, the model from the best epoch was saved in the checkpoint
        callback and is loaded now. If there was not enough training data, the model is only randomly initialized.
        :param gene_expression: gene expression data
        :param mutations: mutation data
        :param copy_number: copy number variation data
        :returns: predicted drug response
        """
        # convert to torch tensors
        gene_expression_tensor = torch.from_numpy(gene_expression).float().to(self.device)
        mutations_tensor = torch.from_numpy(mutations).float().to(self.device)
        copy_number_tensor = torch.from_numpy(copy_number).float().to(self.device)
        self.eval()
        with torch.no_grad():
            z = self._encode_and_concatenate(gene_expression_tensor, mutations_tensor, copy_number_tensor)
            preds = self.regressor(z)
        return preds.squeeze().cpu().detach().numpy()

    def _encode_and_concatenate(
        self, gene_expression: torch.Tensor, mutations: torch.Tensor, copy_number: torch.Tensor
    ) -> torch.Tensor:
        """
        Encodes the input modalities, concatenates, and normalizes the resulting embeddings.

        :param gene_expression: gene expression data
        :param mutations: mutation data
        :param copy_number: copy number variation data
        :returns: concatenated, normalized embeddings
        """
        z_ex = self.expression_encoder(gene_expression)
        z_mu = self.mutation_encoder(mutations)
        z_cn = self.cna_encoder(copy_number)

        z = torch.cat((z_ex, z_mu, z_cn), dim=1)
        z = nn.functional.normalize(z, p=2, dim=0)
        return z

    def forward(self, x_gene: torch.Tensor, x_mutation: torch.Tensor, x_cna: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the MOLIR model.

        :param x_gene: gene expression input
        :param x_mutation: mutation input
        :param x_cna: copy number variation input
        :returns: predicted drug response
        """
        z = self._encode_and_concatenate(x_gene, x_mutation, x_cna)
        preds = self.regressor(z)
        return preds

    def _compute_loss(self, z: torch.Tensor, preds: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Computes the combined triplet loss and regression loss.

        :param z: concatenated, normalized embeddings on which the triplet loss is calculated
        :param preds: predicted drug response on which the regression loss is calculated
        :param y: true drug response
        :returns: combined loss
        """
        positive_indices, negative_indices = generate_triplets_indices(
            y.cpu().detach().numpy(), self.positive_range, self.negative_range
        )

        triplet_loss = self.triplet_loss(z, z[positive_indices], z[negative_indices])
        regression_loss = self.regression_loss(preds.squeeze(), y)
        return triplet_loss + regression_loss

    def training_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Training step of the MOLIR model.

        :param batch: batch of gene expression, mutations, copy number variation, and response
        :param batch_idx: index of the batch
        :returns: combined loss
        """
        gene_expression, mutations, copy_number, response = batch

        # Encode and concatenate
        z = self._encode_and_concatenate(gene_expression, mutations, copy_number)

        # Get predictions
        preds = self.regressor(z)

        # Compute loss
        loss = self._compute_loss(z, preds, response)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        return loss

    def validation_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Validation step of the MOLIR model.

        :param batch: batch of gene expression, mutations, copy number variation, and response
        :param batch_idx: index of the batch
        :returns: combined loss
        """
        gene_expression, mutations, copy_number, response = batch

        # Encode and concatenate
        z = self._encode_and_concatenate(gene_expression, mutations, copy_number)

        # Get predictions
        preds = self.regressor(z)

        # Compute loss
        val_loss = self._compute_loss(z, preds, response)
        self.log("val_loss", val_loss, on_step=False, on_epoch=True, prog_bar=True)
        return val_loss

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """
        Overwrites the configure_optimizers method from PyTorch Lightning.

        :returns: optimizers for the MOLIR expression, mutation, copy number variation encoders, and regressor
        """
        optimizer = torch.optim.Adagrad(
            [
                {"params": self.expression_encoder.parameters(), "lr": self.lr},
                {"params": self.mutation_encoder.parameters(), "lr": self.lr},
                {"params": self.cna_encoder.parameters(), "lr": self.lr},
                {"params": self.regressor.parameters(), "lr": self.lr},
            ],
            weight_decay=self.weight_decay,
        )
        return optimizer



"""Contains the Multi-OMICS Random Forest model."""

import numpy as np
from sklearn.decomposition import PCA

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset

from ..utils import get_multiomics_feature_dataset
from .sklearn_models import RandomForest


class MultiOmicsRandomForest(RandomForest):
    """Multi-OMICS Random Forest model."""

    cell_line_views = [
        "gene_expression",
        "methylation",
        "mutations",
        "copy_number_variation_gistic",
    ]

    def __init__(self):
        """
        Initializes the model.

        Sets the PCA to None, which is initialized in the build_model method.
        """
        super().__init__()
        self.pca = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: MultiOmicsRandomForest
        """
        return "MultiOmicsRandomForest"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Hyperparameters for the model.
        """
        super().build_model(hyperparameters)
        self.pca = PCA(n_components=hyperparameters["n_components"])

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: data path e.g. data/
        :param dataset_name: dataset name e.g. GDSC1
        :returns: FeatureDataset containing the cell line omics features, filtered through the
            drug target genes
        """
        gene_lists = {
            "gene_expression": "drug_target_genes_all_drugs",
            "methylation": "methylation_intersection",
            "mutations": "drug_target_genes_all_drugs",
            "copy_number_variation_gistic": "drug_target_genes_all_drugs",
            "proteomics": "drug_target_genes_all_drugs_proteomics",
        }
        return get_multiomics_feature_dataset(data_path=data_path, gene_lists=gene_lists, dataset_name=dataset_name)

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the model: the number of features is the number of genes + the number of fingerprints.

        :param output: training dataset containing the response output
        :param cell_line_input: training dataset containing the OMICs
        :param drug_input: training dataset containing fingerprints data
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed
        """
        inputs = self.get_feature_matrices(
            cell_line_ids=output.cell_line_ids,
            drug_ids=output.drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        (
            gene_expression,
            methylation,
            mutations,
            copy_number_variation_gistic,
            fingerprints,
        ) = (
            inputs["gene_expression"],
            inputs["methylation"],
            inputs["mutations"],
            inputs["copy_number_variation_gistic"],
            inputs["fingerprints"],
        )

        methylation = self.pca.fit_transform(methylation)

        x = np.concatenate(
            (
                gene_expression,
                methylation,
                mutations,
                copy_number_variation_gistic,
                fingerprints,
            ),
            axis=1,
        )
        self.model.fit(x, output.response)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the response for the given input.

        :param cell_line_ids: cell line ids
        :param drug_ids: drug ids
        :param cell_line_input: cell line input
        :param drug_input: drug input
        :returns: predicted response
        """
        inputs = self.get_feature_matrices(
            cell_line_ids=cell_line_ids,
            drug_ids=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        (
            gene_expression,
            methylation,
            mutations,
            copy_number_variation_gistic,
            fingerprints,
        ) = (
            inputs["gene_expression"],
            inputs["methylation"],
            inputs["mutations"],
            inputs["copy_number_variation_gistic"],
            inputs["fingerprints"],
        )

        methylation = self.pca.transform(methylation)
        x = np.concatenate(
            (
                gene_expression,
                methylation,
                mutations,
                copy_number_variation_gistic,
                fingerprints,
            ),
            axis=1,
        )
        return self.model.predict(x)



"""Draws the critical difference plot.

This method performs the following steps:

1. **Friedman Test**: First, it performs the Friedman test, which is a non-parametric statistical test used to detect
   differences in treatments across multiple test attempts. It compares the ranks of multiple groups and is
   suitable when there are repeated measurements for each group (as is the case here with cross-validation splits).
   The p-value of this test is used to assess whether there are any significant differences in the performance of the
   models.

2. **Post-hoc Conover Test**: If the Friedman test returns a significant result (p-value < 0.05), the post-hoc Conover
   test can be used to identify pairs of algorithms that perform significantly different. This test is necessary
   because the Friedman test only tells if there is a difference somewhere among the models, but not which ones are
   different. The `scikit_posthocs` library is used for this step.

3. **Rank Calculation**: Next, the average ranks of each classifier across all cross-validation splits are computed.
   The models are ranked based on their performance (lower ranks indicate better performance) and the average rank
   across all splits is calculated for each model.

4. **Critical Difference Diagram**: Finally, the method draws the critical difference diagram. This diagram visually
   displays the significant differences between the algorithms. A horizontal line groups a set of models that are
   not significantly different. The critical difference is determined based on the post-hoc test results.
"""

import pathlib
import warnings
from io import TextIOWrapper
from typing import Optional, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.colors as pc
import scikit_posthocs as sp
from matplotlib import pyplot
from matplotlib.axes import Axes
from pandas import DataFrame, Series
from scikit_posthocs import sign_array
from scipy import stats

from ..evaluation import MINIMIZATION_METRICS
from ..pipeline_function import pipeline_function
from .outplot import OutPlot

matplotlib.use("agg")
matplotlib.rcParams["font.family"] = "sans-serif"
matplotlib.rcParams["font.sans-serif"] = "Helvetica Neue"
warnings.filterwarnings("ignore", category=FutureWarning, message=".*swapaxes.*")


class CriticalDifferencePlot(OutPlot):
    """
    Draws the critical difference diagram.

    Used by the pipeline!

    The critical difference diagram is used to compare the performance of multiple classifiers and show whether a
    model is significantly better than another model. This is calculated over the average ranks of the classifiers
    which is why there need to be at least 3 classifiers to draw the diagram. Because the ranks are calculated over
    the cross-validation splits and the significance threshold is set to 0.05, e.g., 10 CV folds are advisable.
    """

    @pipeline_function
    def __init__(self, eval_results_preds: pd.DataFrame, metric="MSE"):
        """
        Initializes the critical difference plot.

        :param eval_results_preds: evaluation results subsetted to predictions only (no randomizations etc)
        :param metric: to be used to assess the critical difference
        """
        eval_results_preds = eval_results_preds[["algorithm", "CV_split", metric]]
        if metric in MINIMIZATION_METRICS:
            eval_results_preds.loc[:, metric] = -eval_results_preds.loc[:, metric]

        self.eval_results_preds = eval_results_preds
        self.metric = metric
        self.fig: Optional[plt.Figure] = None
        self.test_results: Optional[pd.DataFrame] = None

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draws the critical difference plot and saves it to a file.

        :param out_prefix: e.g., results/my_run/critical_difference_plots/
        :param out_suffix: e.g., LPO
        :raises ValueError: if the figure is None or the test results are None
        """
        try:
            self._draw()
            path_out = f"{out_prefix}critical_difference_algorithms_{out_suffix}.svg"
            if self.fig is None or self.test_results is None:
                raise ValueError("Figure is None. Cannot save the plot.")
            else:
                self.fig.savefig(path_out, bbox_inches="tight")
                plt.clf()
                self.test_results = self.test_results.round(4)
                self.test_results.to_html(f"{out_prefix}critical_difference_algorithms_{out_suffix}.html")
        except Exception as e:
            print(f"Error in drawing critical difference plot: {e}")

    def _draw(self) -> None:
        """Draws the critical difference plot."""
        input_friedman = self.eval_results_preds.groupby("algorithm")[self.metric].apply(list)
        # check that all algorithms have the same number of CV splits, if not filter them out
        # table lengths of arrays:
        table_lengths = input_friedman.apply(len)
        # get the most common length
        most_common_length = table_lengths.mode().values[0]
        # filter out algorithms that do not have the most common length
        input_friedman = input_friedman[table_lengths == most_common_length]
        algorithms_included = set(input_friedman.index)
        friedman_p_value = stats.friedmanchisquare(*input_friedman).pvalue
        self.eval_results_preds = self.eval_results_preds[
            self.eval_results_preds["algorithm"].isin(algorithms_included)
        ]
        # transform: rows = CV_split, columns = algorithms, values = metric
        input_conover_friedman = self.eval_results_preds.pivot_table(
            index="CV_split", columns="algorithm", values=self.metric
        )
        self.test_results = pd.DataFrame(sp.posthoc_conover_friedman(input_conover_friedman))
        average_ranks = input_conover_friedman.rank(ascending=False, axis=1).mean(axis=0)
        plt.title(
            f"Critical Difference Diagram: Metric: {self.metric}.\n"
            f"Overall Friedman-Chi2 p-value: {friedman_p_value:.2e}",
            fontsize=20,
        )
        color_palette = dict()
        generated_colors = _generate_discrete_palette(len(input_conover_friedman.columns))
        for alg in input_conover_friedman.columns:
            color_palette[alg] = generated_colors.pop()

        _critical_difference_diagram(ranks=average_ranks, sig_matrix=self.test_results, color_palette=color_palette)

        self.fig = plt.gcf()

    @staticmethod
    def write_to_html(test_mode: str, f: TextIOWrapper, *args, **kwargs) -> TextIOWrapper:
        """
        Inserts the critical difference plot into the HTML report file.

        :param test_mode: test_mode, e.g., LPO
        :param f: HTML report file
        :param args: not needed
        :param kwargs: not needed
        :returns: HTML report file
        """
        path_out_cd = f"critical_difference_plots/critical_difference_algorithms_{test_mode}.svg"
        f.write(f"<object data={path_out_cd}> </object>")
        f.write(
            "<br><br>"
            "This diagram displays the mean rank of each model over all cross-validation splits: Within each CV "
            "split, the models are ranked according to their MSE. We calculate whether a model is significantly "
            "better than another one using the Friedman test and the post-hoc Conover test. "
            "The Friedman test shows whether there are overall differences between the models. After a significant"
            "Friedman test, the pairwise Conover test is performed to identify which models are significantly "
            "outperforming others. One line indicates which models are not significantly different from each "
            "other. The p-values are shown below. This can only be rendered if at least 3 models were run."
        )
        f.write("<br><br>")
        f.write("<h2>Results of Post-Hoc Conover Test</h2>")
        f.write("<br>")
        path_to_table = pathlib.Path(
            pathlib.Path(f.name).parent, f"critical_difference_plots/critical_difference_algorithms_{test_mode}.html"
        )
        if not path_to_table.exists():
            return f
        with open(path_to_table) as conover_results_f:
            conover_results = conover_results_f.readlines()
            conover_results[0] = conover_results[0].replace(
                '<table border="1" class="dataframe">',
                '<table class="display customDataTable" style="width:100%">',
            )
            for line in conover_results:
                f.write(line)
        return f


def _critical_difference_diagram(
    ranks: Union[dict, Series],
    sig_matrix: DataFrame,
    *,
    color_palette: dict,
    ax: Optional[Axes] = None,
    label_fmt_left: str = "{label} ({rank:.2g})",
    label_fmt_right: str = "({rank:.2g}) {label}",
    label_props: Optional[dict] = None,
    marker_props: Optional[dict] = None,
    elbow_props: Optional[dict] = None,
    crossbar_props: Optional[dict] = None,
    text_h_margin: float = 0.01,
    left_only: bool = False,
) -> dict[str, list]:
    """
    Plot a Critical Difference diagram from ranks and post-hoc results.

    :param ranks : dict or Series
        Indicates the rank value for each sample or estimator (as keys or index).

    :param sig_matrix : DataFrame
        The corresponding p-value matrix outputted by post-hoc tests, with
        indices matching the labels in the ranks argument.

    :param ax : matplotlib.SubplotBase, optional
        The object in which the plot will be built. Gets the current Axes
        by default (if None is passed).

    :param label_fmt_left : str, optional
        The format string to apply to the labels on the left side. The keywords
        label and rank can be used to specify the sample/estimator name and
        rank value, respectively, by default '{label} ({rank:.2g})'.

    :param label_fmt_right : str, optional
        The same, but for the labels on the right side of the plot.
        By default '({rank:.2g}) {label}'.

    :param label_props : dict, optional
        Parameters to be passed to pyplot.text() when creating the labels,
        by default None.

    :param marker_props : dict, optional
        Parameters to be passed to pyplot.scatter() when plotting the rank
        markers on the axis, by default None.

    :param elbow_props : dict, optional
        Parameters to be passed to pyplot.plot() when creating the elbow lines,
        by default None.

    :param crossbar_props : dict, optional
        Parameters to be passed to pyplot.plot() when creating the crossbars
        that indicate lack of statistically significant difference. By default
        None.

    :param color_palette: dict
        Parameters to be passed when you need specific colors for each category

    :param text_h_margin : float, optional
        Space between the text labels and the nearest vertical line of an
        elbow, by default 0.01.

    :param left_only: boolean, optional
        Set all labels in a single left-sided block instead of splitting them
        into two block, one for the left and one for the right.
    :raises ValueError: If the color_palette keys are not consistent with the ranks.
    :returns: dict
    """
    # check color_palette consistency
    if isinstance(color_palette, dict) and ((len(set(ranks.keys()) & set(color_palette.keys()))) == len(ranks)):
        pass
    elif isinstance(color_palette, list) and (len(ranks) <= len(color_palette)):
        pass
    else:
        raise ValueError("color_palette keys are not consistent, or list size too small")

    elbow_props = elbow_props or {}
    marker_props = {"zorder": 3, **(marker_props or {})}
    label_props = {"va": "center", "fontsize": 16, "weight": "heavy", **(label_props or {})}
    crossbar_props = {
        "color": "k",
        "zorder": 3,
        "linewidth": 4,
        **(crossbar_props or {}),
    }

    ax = ax or pyplot.gca()
    ax.yaxis.set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.xaxis.set_ticks_position("top")
    ax.spines["top"].set_position("zero")

    # lists of artists to be returned
    markers = []
    elbows = []
    labels = []
    crossbars = []

    # True if pairwise comparison is NOT significant
    adj_matrix = DataFrame(
        1 - sign_array(sig_matrix),
        index=sig_matrix.index,
        columns=sig_matrix.columns,
        dtype=bool,
    )

    ranks = Series(ranks).sort_values()  # Standardize if ranks is dict
    if left_only:
        points_left = ranks
    else:
        left_points = len(ranks) // 2
        points_left, points_right = (
            ranks.iloc[:left_points],
            ranks.iloc[left_points:],
        )

    # for each algorithm: get the set of algorithms that are not significantly different
    crossbar_sets = dict()
    for alg, row in adj_matrix.iterrows():
        not_different = adj_matrix.columns[row].tolist()
        crossbar_sets[alg] = set(not_different).union({alg})

    # Create stacking of crossbars: make a crossbar of the fitting color for each algorithm
    crossbar_levels: list[list[set]] = []
    ypos = -0.5
    for alg in ranks.index:
        bar = crossbar_sets[alg]
        not_different = crossbar_sets[alg]
        if len(not_different) == 1:
            continue
        crossbar_levels.append([bar])

        crossbar_props["color"] = color_palette[alg]
        crossbars.append(
            ax.plot(
                # Adding a separate line between each pair enables showing a
                # marker over each elbow with crossbar_props={'marker': 'o'}.
                [ranks[i] for i in bar],
                [ypos] * len(bar),
                **crossbar_props,
            )
        )
        ypos -= 0.5

    lowest_crossbar_ypos = ypos

    def plot_items(points, xpos, label_fmt, color_palette, label_props):
        """
        Plot each marker + elbow + label.

        :param points: the points to plot
        :param xpos: the x position of the points
        :param label_fmt: the format of the label
        :param color_palette: the color palette to use
        :param label_props: the label properties
        """
        ypos = lowest_crossbar_ypos - 0.5
        for idx, (label, rank) in enumerate(points.items()):
            if not color_palette or len(color_palette) == 0:
                elbow, *_ = ax.plot(
                    [xpos, rank, rank],
                    [ypos, ypos, 0],
                    **elbow_props,
                )
            else:
                elbow, *_ = ax.plot(
                    [xpos, rank, rank],
                    [ypos, ypos, 0],
                    c=color_palette[label] if isinstance(color_palette, dict) else color_palette[idx],
                    **elbow_props,
                )

            elbows.append(elbow)
            curr_color = elbow.get_color()
            markers.append(ax.scatter(rank, 0, **{"color": curr_color, **marker_props}))
            labels.append(
                ax.text(
                    xpos,
                    ypos,
                    label_fmt.format(label=label, rank=rank),
                    color=curr_color,
                    **label_props,
                )
            )
            ypos -= 0.5

    plot_items(
        points_left,
        xpos=points_left.iloc[0] - text_h_margin,
        label_fmt=label_fmt_left,
        color_palette=color_palette,
        label_props={
            "ha": "right",
            **label_props,
        },
    )

    if not left_only:
        plot_items(
            points_right[::-1],
            xpos=points_right.iloc[-1] + text_h_margin,
            label_fmt=label_fmt_right,
            color_palette=color_palette,
            label_props={"ha": "left", **label_props},
        )

    return {
        "markers": markers,
        "elbows": elbows,
        "labels": labels,
        "crossbars": crossbars,
    }


def _generate_discrete_palette(n_colors):
    # Get the base D3 categorical palette
    base_palette = pc.qualitative.D3
    base_n = len(base_palette)  # Number of available discrete colors

    if n_colors <= base_n:
        return base_palette[:n_colors]  # Use available colors directly

    # Convert HEX to RGB (0-1 range)
    base_rgb = np.array([matplotlib.colors.to_rgb(c) for c in base_palette])

    # Generate target indices in the interpolated space
    target_indices = np.linspace(0, base_n - 1, n_colors)

    # Interpolate in RGB space
    interpolated_rgb = np.array(
        [np.interp(target_indices, np.arange(base_n), base_rgb[:, i]) for i in range(3)]
    ).T  # Transpose to get (n_colors, 3)

    # Convert back to HEX
    interpolated_hex = [matplotlib.colors.to_hex(c) for c in interpolated_rgb]

    return interpolated_hex



"""Utility functions for the evaluation pipeline."""

import argparse
from pathlib import Path

from sklearn.base import TransformerMixin
from sklearn.preprocessing import MinMaxScaler, RobustScaler, StandardScaler

from .datasets import AVAILABLE_DATASETS
from .datasets.dataset import DrugResponseDataset
from .datasets.loader import load_dataset
from .datasets.utils import ALLOWED_MEASURES
from .evaluation import AVAILABLE_METRICS
from .experiment import drug_response_experiment, pipeline_function
from .models import MODEL_FACTORY


@pipeline_function
def get_parser() -> argparse.ArgumentParser:
    """
    Get the parser for the evaluation pipeline.

    :returns: parser
    """
    parser = argparse.ArgumentParser(description="Run the drug response prediction model test suite.")
    parser.add_argument(
        "--run_id",
        type=str,
        default="my_run",
        help="identifier to save the results",
    )

    parser.add_argument(
        "--path_data",
        type=str,
        default="data",
        help="Path to the data directory",
    )

    parser.add_argument(
        "--models",
        nargs="+",
        help="model to evaluate or list of models to compare",
    )
    parser.add_argument(
        "--baselines",
        nargs="+",
        help="baseline or list of baselines. The baselines are also hpam-tuned and compared to the "
        "models, but no randomization or robustness tests are run. NaiveMeanEffectsPredictor is always run"
        "as it is required for evaluation",
    )
    parser.add_argument(
        "--test_mode",
        nargs="+",
        default=["LPO"],
        help="Which tests to run (LPO=Leave-random-Pairs-Out, "
        "LCO=Leave-Cell-line-Out, LDO=Leave-Drug-Out). Can be a list of test runs e.g. "
        "'LPO LCO LDO' to run all tests. Default is LPO",
    )
    parser.add_argument(
        "--randomization_mode",
        nargs="+",
        default=["None"],
        help="Which randomization tests to run, additionally to the normal run. Default is None, "
        "which means no randomization tests are run."
        "Modes: SVCC, SVRC, SVCD, SVRD. "
        "Can be a list of randomization tests e.g. 'SCVC SCVD' to run two tests."
        "SVCC: Single View Constant for Cell Lines: in this mode, one experiment is done for every "
        "cell line view the model uses (e.g. gene expression, mutation, ..). "
        "For each experiment one cell line view is held constant while the others are randomized. "
        "SVRC Single View Random for Cell Lines: in this mode, one experiment is done for every "
        "cell line view the model uses (e.g. gene expression, mutation, ..). "
        "For each experiment one cell line view is randomized while the others are held constant. "
        "SVCD: Single View Constant for Drugs: in this mode, one experiment is done for every "
        "drug view the model uses (e.g. fingerprints, target_information, ..). "
        "For each experiment one drug view is held constant while the others are randomized. "
        "SVRD: Single View Random for Drugs: in this mode, one experiment is done for every "
        "drug view the model uses (e.g. gene expression, target_information, ..). "
        "For each experiment one drug view is randomized while the others are held constant.",
    )
    parser.add_argument(
        "--randomization_type",
        type=str,
        default="permutation",
        help='type of randomization to use. Choose from "permutation" or "invariant". Default is '
        '"permutation" "permutation": permute the features over the instances, keeping the '
        "distribution of the features the same but dissolving the relationship to the "
        'target "invariant": the randomization is done in a way that a key characteristic of '
        "the feature is preserved. In case of matrices, this is the mean and standard "
        "deviation of the feature view for this instance, for networks it is the degree "
        "distribution.",
    )
    parser.add_argument(
        "--n_trials_robustness",
        type=int,
        default=0,
        help="Number of trials to run for the robustness test. Default is 0, which means no "
        "robustness test is run. The robustness test is a test where the model is trained "
        "with varying seeds. This is done multiple times to see how stable the model is.",
    )

    parser.add_argument(
        "--dataset_name",
        type=str,
        default="GDSC1",
        help="Name of the drug response dataset",
    )

    parser.add_argument(
        "--cross_study_datasets",
        nargs="+",
        default=[],
        help="List of datasets to use to evaluate predictions across studies. Default is empty list"
        " which means no cross-study datasets are used.",
    )

    parser.add_argument(
        "--path_out",
        type=str,
        default="results/",
        help="Path to the output directory",
    )

    parser.add_argument(
        "--curve_curator",
        action="store_true",
        default=True,
        help="Whether to run " "CurveCurator " "to sort out " "non-reactive " "curves",
    )

    parser.add_argument(
        "--curve_curator_cores",
        type=int,
        default=1,
        help="Max. number of cores used to fit curves with CurveCurator following min(cores, #curves to fit).",
    )

    parser.add_argument(
        "--measure",
        type=str,
        default="LN_IC50",
        help="The drug response measure used as prediction target. Can be one of ['LN_IC50', 'response']",
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="Overwrite existing results with the same path out and run_id? ",
    )
    parser.add_argument(
        "--optim_metric",
        type=str,
        default="RMSE",
        help=f"Metric for hyperparameter tuning choose from {list(AVAILABLE_METRICS.keys())} " f"Default is RMSE.",
    )
    parser.add_argument(
        "--n_cv_splits",
        type=int,
        default=7,
        help="Number of cross-validation splits to use for the evaluation",
    )

    parser.add_argument(
        "--response_transformation",
        type=str,
        default="None",
        help="Transformation to apply to the response variable during training and prediction. "
        "Will be retransformed after the final predictions. Possible values: standard, "
        "minmax, robust",
    )
    parser.add_argument(
        "--multiprocessing",
        action="store_true",
        default=False,
        help="Whether to use multiprocessing for the evaluation. Default is False",
    )
    parser.add_argument(
        "--model_checkpoint_dir",
        type=str,
        default="TEMPORARY",
        help="Directory to save model checkpoints",
    )

    return parser


@pipeline_function
def check_arguments(args) -> None:
    """
    Check the validity of the arguments for the evaluation pipeline.

    :param args: arguments passed from the command line
    :raises AssertionError: if any of the arguments is invalid
    :raises ValueError: if the number of cross-validation splits or curve_curator_cores is less than 1
    :raises FileNotFoundError: if a custom dataset name was specified and the input file could not be found.
    """
    if not args.models:
        raise AssertionError("At least one model must be specified")
    if not all(model in MODEL_FACTORY for model in args.models):
        raise AssertionError(
            f"Invalid model name. Available models are {list(MODEL_FACTORY.keys())}. If you want to "
            f"use your own model, you need to implement a new model class and add it to the "
            f"MODEL_FACTORY in the models init"
        )
    if not all(test in ["LPO", "LCO", "LDO", "LTO"] for test in args.test_mode):
        raise AssertionError("Invalid test mode. Available test modes are LPO, LCO, LDO, LTO")

    if args.baselines is not None:
        if not all(baseline in MODEL_FACTORY for baseline in args.baselines):
            raise AssertionError(
                f"Invalid baseline name. Available baselines are {list(MODEL_FACTORY.keys())}. If you "
                f"want to use your own baseline, you need to implement a new model class and add it to "
                f"the MODEL_FACTORY in the models init"
            )
    if args.dataset_name not in AVAILABLE_DATASETS:
        if args.curve_curator:
            expected_custom_input = Path(args.path_data).absolute() / args.dataset_name / f"{args.dataset_name}_raw.csv"
            if not expected_custom_input.is_file():
                raise FileNotFoundError(
                    "You specified the curve_curator option with a custom dataset name which requires raw "
                    f"viability data to be located at {expected_custom_input} but the file does not exist. "
                    "Please check the 'path_data' and 'dataset_name' arguments and ensure the raw viability "
                    "input file is located at <path_data>/<dataset_name>/<dataset_name>_raw.csv."
                )
        else:
            expected_custom_input = Path(args.path_data).absolute() / args.dataset_name / f"{args.dataset_name}_raw.csv"
            if not expected_custom_input.is_file():
                raise FileNotFoundError(
                    "You specified a custom dataset name which requires prefit curve data to be located at "
                    f"{expected_custom_input} but the file does not exist. Please check the 'path_data' and "
                    "'dataset_name' arguments and ensure the prefit curve data is located at input file is "
                    "located at <path_data>/<dataset_name>/<dataset_name>.csv."
                )

    if args.curve_curator and args.curve_curator_cores < 1:
        raise ValueError("Number of cores for CurveCurator must be greater than 0.")

    for dataset in args.cross_study_datasets:
        if dataset not in AVAILABLE_DATASETS:
            raise AssertionError(
                f"Invalid dataset name in cross_study_datasets. Available datasets are "
                f"{list(AVAILABLE_DATASETS.keys())} If you want to use your own dataset, you "
                f"need to implement a new response dataset loader and add it to the "
                f"AVAILABLE_DATASETS in the response_datasets init."
            )

    # if the path to args.path_data does not exist, create the directory
    Path(args.path_data).mkdir(parents=True, exist_ok=True)

    if args.n_cv_splits <= 1:
        raise ValueError("Number of cross-validation splits must be greater than 1.")

    # TODO Allow for custom randomization tests maybe via config file
    if args.randomization_mode[0] != "None":
        if not all(randomization in ["SVCC", "SVRC", "SVCD", "SVRD"] for randomization in args.randomization_mode):
            raise AssertionError(
                "At least one invalid randomization mode. Available randomization modes are SVCC, " "SVRC, SVSC, SVRD"
            )

    if args.randomization_type not in ["permutation", "invariant"]:
        raise AssertionError("Invalid randomization type. Choose from 'permutation' or 'invariant'")

    if args.n_trials_robustness < 0:
        raise ValueError("Number of trials for robustness test must be greater than or equal to 0")

    if args.measure not in ALLOWED_MEASURES:
        raise ValueError(
            "Only 'LN_IC50', 'EC50', 'IC50', 'pEC50', 'AUC', 'response' or their equivalents including "
            "the '_curvecurator' suffix are allowed drug response measures."
        )

    if args.response_transformation not in ["None", "standard", "minmax", "robust"]:
        raise AssertionError("Invalid response_transformation. Choose from None, standard, minmax, robust")

    if args.optim_metric not in AVAILABLE_METRICS:
        raise AssertionError(
            f"Invalid optim_metric for hyperparameter tuning. Choose from" f" {list(AVAILABLE_METRICS.keys())}"
        )


def main(args, hyperparameter_tuning=True) -> None:
    """
    Main function to run the drug response evaluation pipeline.

    :param args: passed from command line
    :param hyperparameter_tuning: whether to run the pipeline in debug mode
    """
    check_arguments(args)
    # PIPELINE: LOAD_RESPONSE
    response_data, cross_study_datasets = get_datasets(
        dataset_name=args.dataset_name,
        cross_study_datasets=args.cross_study_datasets,
        path_data=args.path_data,
        measure=args.measure,
        curve_curator=args.curve_curator,
        cores=args.curve_curator_cores,
    )

    models = [MODEL_FACTORY[model] for model in args.models]

    if args.baselines is not None:
        baselines = [MODEL_FACTORY[baseline] for baseline in args.baselines]
    else:
        baselines = []

    # NaiveMeanEffectsPredictor is always run as it is needed for evaluation
    if "NaiveMeanEffectsPredictor" not in baselines:
        baselines.append(MODEL_FACTORY["NaiveMeanEffectsPredictor"])

    # TODO Allow for custom randomization tests maybe via config file

    if args.randomization_mode[0] == "None":
        args.randomization_mode = None
    response_transformation = get_response_transformation(args.response_transformation)

    for test_mode in args.test_mode:
        drug_response_experiment(
            models=models,
            baselines=baselines,
            response_data=response_data,
            response_transformation=response_transformation,
            metric=args.optim_metric,
            n_cv_splits=args.n_cv_splits,
            multiprocessing=args.multiprocessing,
            test_mode=test_mode,
            randomization_mode=args.randomization_mode,
            randomization_type=args.randomization_type,
            n_trials_robustness=args.n_trials_robustness,
            cross_study_datasets=cross_study_datasets,
            path_out=args.path_out,
            run_id=args.run_id,
            overwrite=args.overwrite,
            path_data=args.path_data,
            model_checkpoint_dir=args.model_checkpoint_dir,
            hyperparameter_tuning=hyperparameter_tuning,
        )


def get_datasets(
    dataset_name: str,
    cross_study_datasets: list,
    path_data: str = "data",
    measure: str = "response",
    curve_curator: bool = False,
    cores: int = 1,
) -> tuple[DrugResponseDataset, list[DrugResponseDataset] | None]:
    """
    Load the response data and cross-study datasets.

    :param dataset_name: The name of the dataset to load. Can be one of ('GDSC1', 'GDSC2', 'CCLE', CTRPv1',
        'CTRPv2', 'TOYv1', 'TOYv2')
        to download provided datasets, or any other name to use a custom datasets.
    :param cross_study_datasets: list of cross-study datasets. CurveCurator is not applicable to these. If you wish
        to provide custom cross_study_datasets, you have to invoke curve fitting manually using
        drevalpy.datasets.curvecurator.fit_curves
    :param path_data: The parent path in which custom or downloaded datasets should be located, or in which raw
        viability data is to be found for fitting with CurveCurator (see param curve_curator for details).
        The location of the datasets are resolved by <path_data>/<dataset_name>/<dataset_name>.csv.
    :param measure: The name of the column containing the measure to predict, default = "response".
        If curve_curator is True, this measure is appended with "_curvecurator", e.g. "response_curvecurator" to
        distinguish between measures provided by the original source of a dataset, or the measures fit by
        CurveCurator.
    :param curve_curator: If True, the measure is appended with "_curvecurator".
        If a custom dataset_name was provided, this will invoke the fitting procedure of raw viability data,
        which is expected to exist at <path_data>/<dataset_name>/<dataset_name>_raw.csv. The fitted dataset will
        be stored in the same folder, in a file called <dataset_name>.csv
    :param cores: Number of cores to use for CurveCurator fitting. Only used when curve_curator is True, default = 1
    :returns: response data and, potentially, cross-study datasets
    """
    # PIPELINE: LOAD_RESPONSE
    response_data = load_dataset(
        dataset_name=dataset_name, path_data=path_data, measure=measure, curve_curator=curve_curator, cores=cores
    )

    cross_study_datasets = [
        load_dataset(dataset_name=dn, path_data=path_data, measure=measure) for dn in cross_study_datasets
    ]
    return response_data, cross_study_datasets


@pipeline_function
def get_response_transformation(response_transformation: str) -> TransformerMixin | None:
    """
    Get the skelarn response transformation object of choice.

    Users can choose from "None", "standard", "minmax", "robust".
    :param response_transformation: response transformation to apply
    :returns: response transformation object
    :raises ValueError: if the response transformation is not recognized
    """
    if response_transformation == "None":
        return None
    if response_transformation == "standard":
        return StandardScaler()
    if response_transformation == "minmax":
        return MinMaxScaler()
    if response_transformation == "robust":
        return RobustScaler()
    raise ValueError(
        f"Unknown response transformation {response_transformation}. Choose from 'None', "
        f"'standard', 'minmax', 'robust'"
    )



"""Functions for evaluating model performance."""

import numpy as np
from scipy.stats import kendalltau, pearsonr, spearmanr
from sklearn import metrics

from .datasets.dataset import DrugResponseDataset
from .pipeline_function import pipeline_function

warning_shown = False
constant_prediction_warning_shown = False


def _check_constant_prediction(y_pred: np.ndarray) -> bool:
    """
    Check if predictions are constant.

    :param y_pred: predictions
    :return: bool whether predictions are constant
    """
    tol = 1e-6
    # no variation in predictions
    return bool(np.all(np.isclose(y_pred, y_pred[0], atol=tol)))


def _check_constant_target_or_small_sample(y_true: np.ndarray) -> bool:
    """
    Check if target is constant or sample size is too small.

    :param y_true: true response
    :returns: bool whether target is constant or sample size is too small
    """
    tol = 1e-6
    # Check for insufficient sample size or no variation in target
    return len(y_true) < 2 or bool(np.all(np.isclose(y_true, y_true[0], atol=tol)))


def pearson(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    """
    Computes the pearson correlation between predictions and response.

    :param y_pred: predictions
    :param y_true: response
    :return: pearson correlation float
    :raises AssertionError: if predictions and response do not have the same length
    """
    if len(y_pred) != len(y_true):
        raise AssertionError("predictions, response  must have the same length")

    if _check_constant_prediction(y_pred):
        return 0.0
    if _check_constant_target_or_small_sample(y_true):
        return np.nan

    return pearsonr(y_pred, y_true)[0]


def spearman(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    """
    Computes the spearman correlation between predictions and response.

    :param y_pred: predictions
    :param y_true: response
    :return: spearman correlation float
    :raises AssertionError: if predictions and response do not have the same length
    """
    # we can use scipy.stats.spearmanr
    if len(y_pred) != len(y_true):
        raise AssertionError("predictions, response  must have the same length")
    if _check_constant_prediction(y_pred):
        return 0.0
    if _check_constant_target_or_small_sample(y_true):
        return np.nan

    return spearmanr(y_pred, y_true)[0]


def kendall(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    """
    Computes the kendall tau correlation between predictions and response.

    :param y_pred: predictions
    :param y_true: response
    :return: kendall tau correlation float
    :raises AssertionError: if predictions and response do not have the same length
    """
    # we can use scipy.stats.spearmanr
    if len(y_pred) != len(y_true):
        raise AssertionError("predictions, response  must have the same length")
    if _check_constant_prediction(y_pred):
        return 0.0
    if _check_constant_target_or_small_sample(y_true):
        return np.nan

    return kendalltau(y_pred, y_true)[0]


AVAILABLE_METRICS = {
    "MSE": metrics.mean_squared_error,
    "RMSE": metrics.root_mean_squared_error,
    "MAE": metrics.mean_absolute_error,
    "R^2": metrics.r2_score,
    "Pearson": pearson,
    "Spearman": spearman,
    "Kendall": kendall,
}
# both used by pipeline!
MINIMIZATION_METRICS = ["MSE", "RMSE", "MAE"]
MAXIMIZATION_METRICS = ["R^2", "Pearson", "Spearman", "Kendall"]


def get_mode(metric: str):
    """
    Get whether the optimum value of the metric is the minimum or maximum.

    :param metric: metric, e.g., RMSE
    :returns: whether the optimum value of the metric is the minimum or maximum
    :raises ValueError: if the metric is not in MINIMIZATION_METRICS or MAXIMIZATION_METRICS
    """
    if metric in MINIMIZATION_METRICS:
        mode = "min"
    elif metric in MAXIMIZATION_METRICS:
        mode = "max"
    else:
        raise ValueError(
            f"Invalid metric: {metric}. Need to add metric to MINIMIZATION_METRICS or " f"MAXIMIZATION_METRICS?"
        )
    return mode


@pipeline_function
def evaluate(dataset: DrugResponseDataset, metric: list[str] | str):
    """
    Evaluates the model on the given dataset.

    :param dataset: dataset to evaluate on
    :param metric: evaluation metric(s) (one or a list of "MSE", "RMSE", "MAE", "R^2", "Pearson",
        "spearman", "kendall")
    :return: evaluation metric
    :raises AssertionError: if metric is not in AVAILABLE
    """
    if isinstance(metric, str):
        metric = [metric]
    predictions = dataset.predictions
    if predictions is None:
        raise AssertionError("No predictions found in the dataset")
    response = dataset.response

    results = {}
    for m in metric:
        if m not in AVAILABLE_METRICS:
            raise AssertionError(f"invalid metric {m}. Available: {list(AVAILABLE_METRICS.keys())}")
        if len(response) < 2 or np.all(np.isnan(response)) or np.all(np.isnan(predictions)):
            results[m] = float(np.nan)
        else:
            # check whether the predictions contain NaNs
            if np.any(np.isnan(predictions)):
                # if there are only NaNs in the predictions, the metric is NaN
                if np.all(np.isnan(predictions)):
                    results[m] = float(np.nan)
                else:
                    # remove the rows with NaNs in the predictions and response
                    mask = ~np.isnan(predictions)
                    results[m] = float(AVAILABLE_METRICS[m](y_pred=predictions[mask], y_true=response[mask]))
            else:
                results[m] = float(AVAILABLE_METRICS[m](y_pred=predictions, y_true=response))

    return results



"""Gene expression Autoencoder for DIPK model."""

from abc import ABC
from copy import deepcopy

import numpy as np
import torch
import torch.nn
import torch.nn as nn
import torch.optim as optim
from torch.nn import functional
from torch.utils.data import DataLoader, Dataset


class GeneExpressionEncoder(nn.Module):
    """Gene expression encoder.

    Code adapted from the
    DIPK model https://github.com/user15632/DIPK.
    """

    def __init__(self, input_dim, latent_dim=512, h_dims=None, drop_out_rate=0.3):
        """Initialize the gene expression encoder.

        :param input_dim: input dimension
        :param latent_dim: latent dimension
        :param h_dims: hidden dimensions
        :param drop_out_rate: dropout rate
        """
        super().__init__()
        if h_dims is None:
            h_dims = [2048, 1024]
        hidden_dims = deepcopy(h_dims)
        hidden_dims.insert(0, input_dim)
        modules = []
        for i in range(1, len(hidden_dims)):
            modules.append(
                nn.Sequential(
                    nn.Linear(hidden_dims[i - 1], hidden_dims[i]),
                    nn.BatchNorm1d(hidden_dims[i]),
                    nn.ReLU(),
                    nn.Dropout(drop_out_rate),
                )
            )
        self.encoder = nn.Sequential(*modules)
        self.bottleneck = nn.Linear(hidden_dims[-1], latent_dim)
        self.latent_dim = latent_dim

    def forward(self, input):
        """Forward pass of the gene expression encoder.

        :param input: input data
        :return: encoded data
        """
        result = self.encoder(input)
        embedding = functional.relu(self.bottleneck(result))
        return embedding


class GeneExpressionDecoder(nn.Module):
    """Gene expression decoder."""

    def __init__(self, input_dim, latent_dim=512, h_dims=None, drop_out_rate=0.3):
        """Initialize the gene expression decoder.

        :param input_dim: input dimension
        :param latent_dim: latent dimension
        :param h_dims: hidden dimensions
        :param drop_out_rate: dropout rate
        """
        super().__init__()
        if h_dims is None:
            h_dims = [2048, 1024]
        hidden_dims = deepcopy(h_dims)
        hidden_dims.insert(0, input_dim)
        self.decoder_input = nn.Linear(latent_dim, hidden_dims[-1])
        hidden_dims.reverse()
        modules = []
        for i in range(len(hidden_dims) - 2):
            modules.append(
                nn.Sequential(
                    nn.Linear(hidden_dims[i], hidden_dims[i + 1]),
                    nn.BatchNorm1d(hidden_dims[i + 1]),
                    nn.ReLU(),
                    nn.Dropout(drop_out_rate),
                )
            )
        self.decoder = nn.Sequential(*modules)
        self.decoder_output = nn.Linear(hidden_dims[-2], hidden_dims[-1])

    def forward(self, embedding):
        """
        Forward pass of the gene expression decoder.

        :param embedding: input data
        :return: decoded data
        """
        result = self.decoder_input(embedding)
        result = self.decoder(result)
        output = self.decoder_output(result)
        return output


class CollateFn:
    """Collate function for the DataLoader, either for training or testing."""

    def __call__(self, batch):
        """Collate the batch.

        :param batch: batch of PyG Data objects
        :returns: PyG Batch, gene features, and bionic features
        """
        batch_data = torch.stack(batch)
        return batch_data


class DataSet(Dataset, ABC):
    """Dataset class for gene expression data."""

    def __init__(self, data):
        """Initialize the dataset.

        :param data: data
        """
        self._data = data

    def __getitem__(self, idx):
        """Return the data at the given index.

        :param idx: index
        :return: data
        """
        data = self._data[idx]
        return data

    def __len__(self):
        """Return the length of the dataset.

        :return: length of the dataset
        """
        return len(self._data)


def train_gene_expession_autoencoder(
    gene_expression_input: np.ndarray, gene_expression_input_early_stopping: np.ndarray, epochs_autoencoder: int = 100
) -> GeneExpressionEncoder:
    """Train the autoencoder model for gene expression data with early stopping.

    :param gene_expression_input: gene expression data
    :param gene_expression_input_early_stopping: validation data for early stopping
    :param epochs_autoencoder: number of epochs for training the autoencoder
    :return: trained encoder model
    """
    lr = 1e-4
    batch_size = 1024
    noising = True
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # create model
    encoder = GeneExpressionEncoder(len(gene_expression_input[0])).to(device)
    decoder = GeneExpressionDecoder(len(gene_expression_input[0])).to(device)
    loss_func = nn.MSELoss()
    params = [{"params": encoder.parameters()}, {"params": decoder.parameters()}]
    optimizer = optim.Adam(params, lr=lr)

    # load data
    my_collate = CollateFn()
    gene_expression_tensor = torch.tensor(gene_expression_input, dtype=torch.float32).to(device)
    train_loader = DataLoader(
        DataSet(gene_expression_tensor), batch_size=batch_size, shuffle=True, collate_fn=my_collate
    )

    # prepare early stopping validation data
    gene_expression_val_tensor = torch.tensor(gene_expression_input_early_stopping, dtype=torch.float32).to(device)

    # early stopping parameters
    patience = 3
    best_val_loss = float("inf")
    epochs_without_improvement = 0

    print("Training DIPK autoencoder for gene expression data")
    for epoch_index in range(epochs_autoencoder):
        # training
        encoder.train()
        decoder.train()
        epoch_loss = 0.0
        batch_count = 0
        for gene_expression_batch in train_loader:
            gene_expression_batch = gene_expression_batch.to(device)
            if noising:
                z = gene_expression_batch.clone()
                y = np.random.binomial(1, 0.2, (z.shape[0], z.shape[1]))
                z[np.array(y, dtype=bool)] = 0
                gene_expression_batch.requires_grad_(True)
                output = decoder(encoder(z))
            else:
                output = decoder(encoder(gene_expression_batch))
            loss = loss_func(output, gene_expression_batch)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            epoch_loss += loss.detach().item()
            batch_count += 1
        epoch_loss /= batch_count

        # validation
        encoder.eval()
        decoder.eval()
        with torch.no_grad():
            val_output = decoder(encoder(gene_expression_val_tensor))
            val_loss = loss_func(val_output, gene_expression_val_tensor).item()

        print(f"DIPK Autoenc. Epoch: {epoch_index}, Train Loss: {epoch_loss}, Val Loss: {val_loss}")

        # early stopping check
        if val_loss < best_val_loss:
            best_val_loss = val_loss
            epochs_without_improvement = 0
        else:
            epochs_without_improvement += 1
            if epochs_without_improvement >= patience:
                print(f"DIPK Autoenc. Early stopping triggered at epoch {epoch_index}")
                break

    encoder.eval()
    return encoder


def encode_gene_expression(gene_expression_input: np.ndarray, encoder: GeneExpressionEncoder) -> np.ndarray:
    """Encode gene expression data.

    :param gene_expression_input: gene expression data
    :param encoder: trained encoder model
    :return: encoded gene expression data
    """
    encoder.eval()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    encoder.to(device)

    # Check the original input shape, because we have to unsqueeze the input if it is a single vector
    original_shape = gene_expression_input.shape

    gene_expression_tensor = torch.tensor(gene_expression_input, dtype=torch.float32).to(device)

    # Add batch dimension if input is a single vector
    if gene_expression_tensor.ndim == 1:
        gene_expression_tensor = gene_expression_tensor.unsqueeze(0)

    with torch.no_grad():
        encoded_data = encoder(gene_expression_tensor).cpu().numpy()

    # Match the output shape to the input shape
    if len(original_shape) == 1:
        encoded_data = encoded_data.squeeze(0)

    return encoded_data



"""Module containing the drevalpy plotly visualizations."""

__all__ = [
    "ComparisonScatter",
    "CriticalDifferencePlot",
    "Heatmap",
    "RegressionSliderPlot",
    "VioHeat",
    "Violin",
    "CrossStudyTables",
]

from .comp_scatter import ComparisonScatter
from .critical_difference_plot import CriticalDifferencePlot
from .cross_study_tables import CrossStudyTables
from .heatmap import Heatmap
from .regression_slider_plot import RegressionSliderPlot
from .vioheat import VioHeat
from .violin import Violin



"""Decorator to mark a function as a pipeline function."""


def pipeline_function(func):
    """
    Decorator to mark a function as a pipeline function.

    :param func: function to decorate
    :return: function with custom attribute
    """
    func.is_pipeline_function = True  # Adds a custom attribute to the function
    return func



"""Utility functions for loading and processing data."""

import os.path

import numpy as np
import pandas as pd

from drevalpy.datasets.dataset import FeatureDataset
from drevalpy.datasets.utils import CELL_LINE_IDENTIFIER, DRUG_IDENTIFIER, TISSUE_IDENTIFIER


def load_cl_ids_from_csv(path: str, dataset_name: str) -> FeatureDataset:
    """
    Load cell line ids from csv file.

    :param path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :returns: FeatureDataset with the cell line ids
    """
    cl_names = pd.read_csv(f"{path}/{dataset_name}/cell_line_names.csv", index_col=1)
    return FeatureDataset(features={cl: {CELL_LINE_IDENTIFIER: np.array([cl])} for cl in cl_names.index})


def load_tissues_from_csv(path: str, dataset_name: str) -> FeatureDataset:
    """
    Load tissues from csv file.

    :param path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :returns: FeatureDataset with the tissues
    """
    tissues = pd.read_csv(f"{path}/{dataset_name}/cell_line_names.csv", index_col=1).drop_duplicates()
    return FeatureDataset(
        features={cl: {TISSUE_IDENTIFIER: np.array([tissues.loc[cl, "tissue"]])} for cl in tissues.index}
    )


def load_and_select_gene_features(
    feature_type: str,
    gene_list: str | None,
    data_path: str,
    dataset_name: str,
) -> FeatureDataset:
    """
    Load and reduce features of a single feature type, ensuring selection and ordering based on the gene list.

    Attention: if gene_list is None, all features are loaded, which can be problematic for cross study prediction.

    :param feature_type: type of feature, e.g., gene_expression, methylation, etc.
    :param gene_list: list of genes to include, e.g., landmark_genes
    :param data_path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :returns: FeatureDataset with the reduced features
    :raises ValueError: if genes from gene_list are missing in the dataset
    """
    ge = pd.read_csv(f"{data_path}/{dataset_name}/{feature_type}.csv", index_col=1)
    ge = ge.drop(columns=["cellosaurus_id"])

    cl_features = FeatureDataset(
        features=iterate_features(df=ge, feature_type=feature_type),
        meta_info={feature_type: ge.columns.values},
    )

    if gene_list is None:
        return cl_features

    gene_info = pd.read_csv(
        f"{data_path}/{dataset_name}/gene_lists/{gene_list}.csv",
        sep=",",
    )
    ordered_genes = gene_info["Symbol"].tolist()

    genes_in_features = set(cl_features.meta_info[feature_type])
    missing_genes = [gene for gene in ordered_genes if gene not in genes_in_features]

    if missing_genes:
        missing_str = (
            f"{', '.join(missing_genes[:10])}, ... ({len(missing_genes)} genes in total)"
            if len(missing_genes) > 10
            else ", ".join(missing_genes)
        )
        raise ValueError(
            f"The following genes are missing from the dataset {dataset_name} for {feature_type}: {missing_str}"
        )

    indices_to_keep = [i for i, gene in enumerate(cl_features.meta_info[feature_type]) if gene in ordered_genes]

    cl_features.meta_info[feature_type] = np.array(ordered_genes)

    for cell_line in cl_features.features.keys():
        cl_features.features[cell_line][feature_type] = cl_features.features[cell_line][feature_type][indices_to_keep]

    return cl_features


def iterate_features(df: pd.DataFrame, feature_type: str) -> dict[str, dict[str, np.ndarray]]:
    """
    Iterate over features.

    :param df: DataFrame with the features
    :param feature_type: type of feature, e.g., gene_expression, methylation, etc.
    :returns: dictionary with the features
    """
    features: dict[str, dict[str, np.ndarray]] = {}
    for cl in df.index:
        if cl in features.keys():
            continue
        rows = df.loc[cl]
        rows = rows.astype(float).to_numpy()
        if (len(rows.shape) > 1) and (rows.shape[0] > 1):  # multiple rows returned
            # take mean
            rows = np.mean(rows, axis=0)
        features[cl] = {feature_type: rows}
    return features


def load_drug_ids_from_csv(data_path: str, dataset_name: str) -> FeatureDataset:
    """
    Load drug ids from csv file.

    :param data_path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :returns: FeatureDataset with the drug ids
    """
    drug_names = pd.read_csv(f"{data_path}/{dataset_name}/drug_names.csv", index_col=0)
    drug_names.index = drug_names.index.astype(str)
    return FeatureDataset(features={drug: {DRUG_IDENTIFIER: np.array([drug])} for drug in drug_names.index})


def load_drug_fingerprint_features(data_path: str, dataset_name: str, fill_na=True, n_bits=128) -> FeatureDataset:
    """
    Load drug features from fingerprints.

    :param data_path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :param fill_na: whether to use default pubchemid-hashed fingerprints if fingerprint is not available
    :param n_bits: number of bits in the fingerprint
    :returns: FeatureDataset with the drug fingerprints
    """
    fingerprints = pd.read_csv(
        os.path.join(data_path, dataset_name, "drug_fingerprints", f"pubchem_id_to_demorgan_{n_bits}_map.csv"),
        index_col=None,
    ).T
    if fill_na:
        for drug in fingerprints.index:
            if (
                not fingerprints.loc[drug].isna().all()
            ):  # if all values are NaN, replace with random fingerprint for the drug
                continue
            # Create random fingerprint for the drug, which is based on a hash of the pubchemid
            rng = np.random.default_rng(hash(drug) % (2**32))
            fingerprints.loc[drug] = rng.integers(0, 2, size=fingerprints.loc[drug].shape)

    return FeatureDataset(
        features={drug: {"fingerprints": fingerprints.loc[drug].values} for drug in fingerprints.index}
    )


def get_multiomics_feature_dataset(
    data_path: str,
    dataset_name: str,
    gene_lists: dict | None = None,
    omics: list[str] | None = None,
) -> FeatureDataset:
    """
    Get multiomics feature dataset for the given list of OMICs.

    :param data_path: path to the data, e.g., data/
    :param dataset_name: name of the dataset, e.g., GDSC2
    :param gene_lists: dictionary of names of lists of genes to include, for each omics type,
                e.g., {"gene_expression": "landmark_genes"}, if None, all features are not reduced
    :param omics: list of omics to include, e.g., ["gene_expression", "methylation"]
    :returns: FeatureDataset with the multiomics features
    :raises ValueError: if no omics features are found
    """
    if omics is None:
        omics = ["gene_expression", "methylation", "mutations", "copy_number_variation_gistic", "proteomics"]

    if gene_lists is None:
        gene_lists = {o: None for o in omics}

    if not np.all([k in omics for k in gene_lists.keys()]):
        raise ValueError("Gene lists must be provided for all omics types.")

    feature_dataset = None
    for omic in omics:
        if feature_dataset is None:
            feature_dataset = load_and_select_gene_features(
                feature_type=omic,
                gene_list=gene_lists[omic],
                data_path=data_path,
                dataset_name=dataset_name,
            )
        else:
            feature_dataset.add_features(
                load_and_select_gene_features(
                    feature_type=omic,
                    gene_list=gene_lists[omic],
                    data_path=data_path,
                    dataset_name=dataset_name,
                )
            )
    if feature_dataset is None:
        raise ValueError("No omics features found.")
    return feature_dataset


def unique(array):
    """
    Get unique values ordered by first occurrence.

    :param array: array of values
    :returns: unique values ordered by first occurrence
    """
    uniq, index = np.unique(array, return_index=True)
    return uniq[index.argsort()]



"""
Contains the SuperFELTR model.

Regression extension of Super.FELT: supervised feature extraction learning using triplet loss for drug response
prediction with multi-omics data.
Very similar to MOLI. Differences:

    * In MOLI, encoders and the classifier were trained jointly. Super.FELT trains them independently
    * MOLI was trained without feature selection (except for the Variance Threshold on the gene expression).
        Super.FELT uses feature selection for all omics data.

The input remains the same: somatic mutation, copy number variation and gene expression data.
Original authors of SuperFELT: Park, Soh & Lee. (2021, 10.1186/s12859-021-04146-z)
Code adapted from their Github: https://github.com/DMCB-GIST/Super.FELT
and Hauptmann et al. (2023, 10.1186/s12859-023-05166-7) https://github.com/kramerlab/Multi-Omics_analysis
"""

from typing import Any

import numpy as np
import pytorch_lightning as pl

from ...datasets.dataset import DrugResponseDataset, FeatureDataset
from ..drp_model import DRPModel
from ..MOLIR.utils import filter_and_sort_omics, get_dimensions_of_omics_data, make_ranges, select_features_for_view
from ..utils import get_multiomics_feature_dataset
from .utils import SuperFELTEncoder, SuperFELTRegressor, train_superfeltr_model


class SuperFELTR(DRPModel):
    """Regression extension of Super.FELT."""

    is_single_drug_model = True
    cell_line_views = ["gene_expression", "mutations", "copy_number_variation_gistic"]
    drug_views = []
    early_stopping = True

    def __init__(self) -> None:
        """
        Initialization method for SuperFELTR Model.

        The encoders and the regressor are initialized to None because they are built later in the first training pass.
        The hyperparameters are also initialized to an empty dict because they are initialized in build_model. The
        ranges are initialized during training which is why here, they get dummy values. The best checkpoint is
        determined after training.
        """
        super().__init__()
        # encoders and regressor are initialized to None because they are built later in the first training pass
        self.expr_encoder: SuperFELTEncoder | None = None
        self.mut_encoder: SuperFELTEncoder | None = None
        self.cnv_encoder: SuperFELTEncoder | None = None
        self.regressor: SuperFELTRegressor | None = None
        # hyperparameters are initialized to None because they are initialized in build_model
        self.hyperparameters: dict[str, Any] = dict()
        # ranges are initialized later because they are initialized using the standard variation of the train
        # response data which is only available when entering the training
        self.ranges: tuple[float, float] = (0.0, 1.0)
        # best checkpoint is determined after training
        self.best_checkpoint: pl.callbacks.ModelCheckpoint | None = None
        self.gene_expression_features = None
        self.mutations_features = None
        self.copy_number_variation_features = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: SuperFELTR
        """
        return "SuperFELTR"

    def build_model(self, hyperparameters) -> None:
        """
        Builds the model from hyperparameters.

        :param hyperparameters: dictionary containing the hyperparameters for the model. Contain mini_batch,
            dropout_rate, weight_decay, out_dim_expr_encoder, out_dim_mutation_encoder, out_dim_cnv_encoder, epochs,
            variance thresholds for gene expression, mutation, and copy number variation, margin, and learning rate.
        """
        self.hyperparameters = hyperparameters

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "superfeltr_checkpoints",
    ) -> None:
        """
        Does feature selection, trains the encoders sequentially, and then trains the regressor.

        If there is not enough training data, the model is trained with random initialization, if there is no
        training data at all, the model is skipped and later on, NA is predicted.

        :param output: training data associated with the response output
        :param cell_line_input: cell line omics features
        :param drug_input: not needed, as it is a single drug model
        :param output_earlystopping: optional early stopping dataset
        :param model_checkpoint_dir: not needed
        :raises ValueError: if drug_input is not None
        """
        if drug_input is not None:
            raise ValueError("SuperFELTR is a single drug model and does not require drug input.")

        if len(output) > 0:
            cell_line_input = self._feature_selection(output, cell_line_input)
            if output_earlystopping is not None and self.early_stopping and len(output_earlystopping) < 2:
                output_earlystopping = None
            dim_gex, dim_mut, dim_cnv = get_dimensions_of_omics_data(cell_line_input)
            self.ranges = make_ranges(output)

            # difference to MOLI: encoders and regressor are trained independently
            # Create and train encoders
            encoders = {}
            encoder_dims = {"expression": dim_gex, "mutation": dim_mut, "copy_number_variation_gistic": dim_cnv}
            for omic_type, dim in encoder_dims.items():
                encoder = SuperFELTEncoder(
                    input_size=dim, hpams=self.hyperparameters, omic_type=omic_type, ranges=self.ranges
                )
                if len(output) >= self.hyperparameters["mini_batch"]:
                    print(f"Training SuperFELTR Encoder for {omic_type} ... ")
                    best_checkpoint = train_superfeltr_model(
                        model=encoder,
                        hpams=self.hyperparameters,
                        output_train=output,
                        cell_line_input=cell_line_input,
                        output_earlystopping=output_earlystopping,
                        patience=5,
                        model_checkpoint_dir=model_checkpoint_dir,
                    )
                    encoders[omic_type] = SuperFELTEncoder.load_from_checkpoint(best_checkpoint.best_model_path)
                else:
                    print(
                        f"Not enough training data provided for SuperFELTR Encoder for {omic_type}. Using random "
                        f"initialization."
                    )
                    encoders[omic_type] = encoder

            self.expr_encoder, self.mut_encoder, self.cnv_encoder = (
                encoders["expression"],
                encoders["mutation"],
                encoders["copy_number_variation_gistic"],
            )

            self.regressor = SuperFELTRegressor(
                input_size=self.hyperparameters["out_dim_expr_encoder"]
                + self.hyperparameters["out_dim_mutation_encoder"]
                + self.hyperparameters["out_dim_cnv_encoder"],
                hpams=self.hyperparameters,
                encoders=(self.expr_encoder, self.mut_encoder, self.cnv_encoder),
            )
            if len(output) >= self.hyperparameters["mini_batch"]:
                print("Training SuperFELTR Regressor ... ")
                self.best_checkpoint = train_superfeltr_model(
                    model=self.regressor,
                    hpams=self.hyperparameters,
                    output_train=output,
                    cell_line_input=cell_line_input,
                    output_earlystopping=output_earlystopping,
                    patience=5,
                    model_checkpoint_dir=model_checkpoint_dir,
                )
            else:
                print("Not enough training data provided for SuperFELTR Regressor. Using random initialization.")
                self.best_checkpoint = None
        else:
            print("No training data provided, skipping model")
            self.best_checkpoint = None
            self.expr_encoder, self.mut_encoder, self.cnv_encoder, self.regressor = None, None, None, None
        if self.best_checkpoint is not None:
            # load best model
            self.regressor = SuperFELTRegressor.load_from_checkpoint(
                self.best_checkpoint.best_model_path,
                input_size=self.hyperparameters["out_dim_expr_encoder"]
                + self.hyperparameters["out_dim_mutation_encoder"]
                + self.hyperparameters["out_dim_cnv_encoder"],
                hpams=self.hyperparameters,
                encoders=(self.expr_encoder, self.mut_encoder, self.cnv_encoder),
            )

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the drug response.

        If there is no training data, NA is predicted. If there was not enough training data, predictions are made
        with the randomly initialized model.

        :param cell_line_ids: cell line ids
        :param drug_ids: drug ids
        :param cell_line_input: cell line omics features
        :param drug_input: drug omics features, not needed
        :returns: predicted drug response
        :raises ValueError: if drug_input is not None
        """
        if self.expr_encoder is None or self.mut_encoder is None or self.cnv_encoder is None or self.regressor is None:
            print("No training data was available, predicting NA")
            return np.array([np.nan] * len(cell_line_ids))
        if (
            self.gene_expression_features is None
            or self.mutations_features is None
            or self.copy_number_variation_features is None
        ):
            raise ValueError("Model was not trained, no features available.")

        if drug_input is not None:
            raise ValueError("SuperFELTR is a single drug model and does not require drug input.")

        input_data = self.get_feature_matrices(
            cell_line_ids=cell_line_ids,
            drug_ids=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        gene_expression, mutations, cnvs = (
            input_data["gene_expression"],
            input_data["mutations"],
            input_data["copy_number_variation_gistic"],
        )

        (gene_expression, mutations, cnvs) = filter_and_sort_omics(
            model=self, gene_expression=gene_expression, mutations=mutations, cnvs=cnvs, cell_line_input=cell_line_input
        )

        if self.best_checkpoint is None:
            print("Not enough training data provided for SuperFELTR Regressor. Predicting with random initialization.")
            return self.regressor.predict(gene_expression, mutations, cnvs)

        return self.regressor.predict(gene_expression, mutations, cnvs)

    def _feature_selection(self, output: DrugResponseDataset, cell_line_input: FeatureDataset) -> FeatureDataset:
        """
        Feature selection for all omics data.

        Originally, this was done with VarianceThreshold but as data can vary and hence the thresholds are not
        universally applicable, we now changed it to select the top 1000 variable features for each omics data.

        :param output: training data associated with the response output
        :param cell_line_input: cell line omics features
        :returns: cell line omics features with selected features
        """
        for view in self.cell_line_views:
            cell_line_input = select_features_for_view(
                view=view,
                cell_line_input=cell_line_input,
                output=output,
            )
        self.gene_expression_features = cell_line_input.meta_info["gene_expression"]
        self.mutations_features = cell_line_input.meta_info["mutations"]
        self.copy_number_variation_features = cell_line_input.meta_info["copy_number_variation_gistic"]
        return cell_line_input

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features: gene expression, mutations, and copy number variation.

        :param data_path: path to the data, e.g., data/
        :param dataset_name: name of the dataset, e.g., GDSC2
        :returns: FeatureDataset containing the cell line gene expression features, mutations, and copy number variation
        """
        feature_dataset = get_multiomics_feature_dataset(
            data_path=data_path, dataset_name=dataset_name, gene_lists=None, omics=self.cell_line_views
        )
        # log transformation
        feature_dataset.apply(function=np.arcsinh, view="gene_expression")
        return feature_dataset

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset | None:
        """
        Returns None, as drug features are not needed for SuperFELTR.

        :param data_path: Path to the fingerprints, e.g., data/
        :param dataset_name: Name of the dataset
        :returns: None
        """
        return None
